#include "Configuration.h"
#include "Configuration.tab.cc"

/******************************************************************************
 *
 * Function definitions for class Configuration.
 *
 *****************************************************************************/

/* ========================================================================= */
Configuration::Configuration()
/* ----------------------------------------------------------------------------
 *
 * Description:   Constructor for class Configuration. Creates a new data
 *                structure for storing program configuration and initializes
 *                the configuration parameters to their default values.
 *
 * Arguments:     None.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  reset();
}

/* ========================================================================= */
Configuration::~Configuration()
/* ----------------------------------------------------------------------------
 *
 * Description:   Destructor for class Configuration.
 *
 * Arguments:     None.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  for (vector<AlgorithmInformation, ALLOC(AlgorithmInformation) >
	 ::const_iterator it = algorithms.begin();
       it != algorithms.end(); ++it)
  {
    delete it->name;
    delete it->path_to_program;
    delete it->extra_parameters;
  }
}

/* ========================================================================= */
void Configuration::read()
/* ----------------------------------------------------------------------------
 *
 * Description:   Reads the program configuration from the file `config'
 *                located in the current directory.
 *
 * Arguments:     None.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  reset();

  if ((yyin = fopen("config", "r")) == NULL)
    throw ConfigurationException(-1, "Error opening `config'");

  /*
   *  Initialize the default operator priorities.
   */

  default_operator_priority = 10;
  symbol_priority[RandomFormulaParameters::ATOM] = 90;
  symbol_priority[RandomFormulaParameters::TRUE] = 5;
  symbol_priority[RandomFormulaParameters::FALSE] = 5;
  symbol_priority[RandomFormulaParameters::AND] = -1;
  symbol_priority[RandomFormulaParameters::OR] = -1;
  symbol_priority[RandomFormulaParameters::UNTIL] = -1;
  symbol_priority[RandomFormulaParameters::V] = -1;
  symbol_priority[RandomFormulaParameters::IMPLICATION] = -1;
  symbol_priority[RandomFormulaParameters::EQUIVALENCE] = -1;
  symbol_priority[RandomFormulaParameters::NOT] = -1;
  symbol_priority[RandomFormulaParameters::NEXT] = -1;
  symbol_priority[RandomFormulaParameters::FINALLY] = -1;
  symbol_priority[RandomFormulaParameters::GLOBALLY] = -1;

  /*
   *  Read the configuration.
   */

  try
  {
    yyparse();
    fclose(yyin);
  }
  catch (ConfigurationException)
  {
    fclose(yyin);
    throw;
  }

  /*
   *  Verify that there is at least one algorithm available for use.
   */

  if (algorithms.empty())
    throw ConfigurationException
            (-1, "No algorithms defined in the configuration file.");

  /*
   *  The case where the number of available variables for propositional
   *  formulae is zero is equivalent to the case where propositional atoms are
   *  disallowed in the formulae altogether.
   */

  if (formula_options.generation_parameters.number_of_available_variables == 0)
    symbol_priority[RandomFormulaParameters::ATOM] = 0;

  /*
   *  Verify that at least one propositional symbol class (a Boolean constant
   *  or a propositional atom) has a nonzero priority.
   */

  if (symbol_priority[RandomFormulaParameters::ATOM] == 0
      && symbol_priority[RandomFormulaParameters::TRUE] == 0
      && symbol_priority[RandomFormulaParameters::FALSE] == 0)
    throw ConfigurationException(-1, "At least one propositional symbol "
                                     "must have nonzero priority.");

  /*
   *  If the operators ->, <->, <> and [] are disallowed, set their priorities 
   *  to zero.
   */

  if (!formula_options.allow_abbreviated_operators)
  {
    symbol_priority[RandomFormulaParameters::IMPLICATION] = 0;
    symbol_priority[RandomFormulaParameters::EQUIVALENCE] = 0;
    symbol_priority[RandomFormulaParameters::FINALLY] = 0;
    symbol_priority[RandomFormulaParameters::GLOBALLY] = 0;
  }

  /*
   *  Verify that at least one unary and one binary operator have a nonzero
   *  priority. Initialize also the priorities of the operators with the
   *  `default' priority.
   */

  bool unary_operator_allowed = false, binary_operator_allowed = false;

  for (map<int, int, less<int>, ALLOC(int) >::iterator
	 it = symbol_priority.begin();
       it != symbol_priority.end(); ++it)
  {
    if (it->second == -1)
      it->second = default_operator_priority;

    if (it->second > 0)
    {
      if (!unary_operator_allowed)
        unary_operator_allowed = 
          (it->first == RandomFormulaParameters::NOT 
           || it->first == RandomFormulaParameters::NEXT 
           || it->first == RandomFormulaParameters::FINALLY
           || it->first == RandomFormulaParameters::GLOBALLY);
      if (!binary_operator_allowed)
        binary_operator_allowed = 
          (it->first == RandomFormulaParameters::AND 
           || it->first == RandomFormulaParameters::OR 
           || it->first == RandomFormulaParameters::UNTIL
           || it->first == RandomFormulaParameters::V
           || it->first == RandomFormulaParameters::IMPLICATION
           || it->first == RandomFormulaParameters::EQUIVALENCE);
    }
  }

  if (!unary_operator_allowed || !binary_operator_allowed)
    throw ConfigurationException(-1, "At least one unary and one binary "
                                     "operator must have nonzero priority.");

  /*
   *  Collect the allowed symbols and operators with their priorities into
   *  their respective symbol lists from which the symbols are randomly
   *  selected when generating formulae.
   */

  for (map<int, int, less<int>, ALLOC(int) >::iterator
	 it = symbol_priority.begin();
       it != symbol_priority.end(); ++it)
  {
    if (it->second != 0)
    {
      switch (it->first)
      {
        case RandomFormulaParameters::ATOM :
        case RandomFormulaParameters::TRUE :
        case RandomFormulaParameters::FALSE :
          formula_options.generation_parameters.usePropositionalSymbol(*it);
          break;

        case RandomFormulaParameters::NOT :
          formula_options.generation_parameters.useShortFormulaOperator(*it);
          if (formula_options.generate_mode != NNF)
            formula_options.generation_parameters.useLongFormulaOperator(*it);
          break;

        case RandomFormulaParameters::NEXT :
          formula_options.generation_parameters.useShortFormulaOperator(*it);
          formula_options.generation_parameters.useLongFormulaOperator(*it);
          break;

        case RandomFormulaParameters::AND :
        case RandomFormulaParameters::OR :
        case RandomFormulaParameters::UNTIL :
        case RandomFormulaParameters::V :
          formula_options.generation_parameters.useLongFormulaOperator(*it);
          break;

        case RandomFormulaParameters::FINALLY :
        case RandomFormulaParameters::GLOBALLY :
          if (formula_options.allow_abbreviated_operators)
          {
            formula_options.generation_parameters.useShortFormulaOperator(*it);
            formula_options.generation_parameters.useLongFormulaOperator(*it);
          }
          break;

        case RandomFormulaParameters::IMPLICATION :
        case RandomFormulaParameters::EQUIVALENCE :
          if (formula_options.allow_abbreviated_operators)
            formula_options.generation_parameters.useLongFormulaOperator(*it);
          break;
      }
    }
  }
}

/* ========================================================================= */
void Configuration::print(ostream& stream = cout, int indent = 0) const
/* ----------------------------------------------------------------------------
 *
 * Description:   Writes information about the program configuration into a
 *                stream.
 *
 * Argument:      stream  --  A reference to an output stream.
 *                indent  --  Number of spaces to leave to the left of output.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  Exceptional_ostream estream(&stream, ios::badbit | ios::failbit);

  estream << string(indent, ' ') + "Program configuration:\n"
             + string(indent + 2, ' ')
          << configuration.global_options.number_of_rounds
          << string(" test round")
             + (configuration.global_options.number_of_rounds > 1 ? "s" : "")
             + ".\n";

  if (configuration.global_options.interactive == ALWAYS)
    estream << string(indent + 2, ' ') + "Pausing after every test round.\n";
  else if (configuration.global_options.interactive == ONERROR)
    estream << string(indent + 2, ' ')
               + "Testing will be interrupted in case of an error.\n";

  estream << string(indent + 2, ' ')
             + "Computing "
             + (configuration.global_options.product_mode == GLOBAL
		? "full" : "restricted")
             + " synchronous products.\n\n" + string(indent + 2, ' ')
             + "Algorithms:\n";

  vector<AlgorithmInformation, ALLOC(AlgorithmInformation) >::size_type
    algorithm_number = 0;
  
  for (vector<AlgorithmInformation, ALLOC(AlgorithmInformation) >
	 ::const_iterator a = algorithms.begin();
       a != algorithms.end();
       ++a)
  {
    estream << string(indent + 4, ' ')
            << algorithm_number
            << string(": `") + *(a->name) + '\'';

    if (!a->enabled)
      estream << " (disabled)";

    estream << '\n';

    algorithm_number++;
  }

  estream << endl
          << string(indent + 2, ' ') + "Random state spaces:\n"
             + string(indent + 4, ' ');

  switch (statespace_options.generate_mode)
  {
    case RANDOMGRAPH :
      estream << "Random graphs ";
      break;
    case RANDOMCONNECTEDGRAPH :
      estream << "Random connected graphs ";
      break;
    case RANDOMPATH :
      estream << "Random paths ";
      break;
    case ENUMERATEDPATH :
      estream << "Enumerated paths ";
      break;
    default :
      break;
  }

  estream << '('
	  << statespace_options.size;

  if (statespace_options.max_size != statespace_options.size)
    estream << "..." << statespace_options.max_size;

  estream << string(" state") + (statespace_options.max_size == 1 ? "" : "s")
             + ", "
          << statespace_options.atoms_per_state
          << string(" atomic proposition")
             + (statespace_options.atoms_per_state == 1 ? "" : "s")
             + ")\n" + string(indent + 4, ' ');

  if (global_options.statespace_change_interval == 0)
    estream << "Using a fixed state space.\n" + string(indent + 4, ' ');
  else
  {
    estream << "New state space will be generated after every ";
    if (global_options.statespace_change_interval > 1)
    {
      estream << global_options.statespace_change_interval;

      if (global_options.statespace_change_interval < 10
	  || global_options.statespace_change_interval > 20)
      {
	switch (global_options.statespace_change_interval % 10)
	{
	  case 1 : estream << "st"; break;
	  case 2 : estream << "nd"; break;
	  case 3 : estream << "rd"; break;
	  default : estream << "th"; break;
	}
      }
      else
	estream << "th";

      estream << ' ';
    }

    estream << "round.\n" + string(indent + 4, ' ');
  }

  if (statespace_options.generate_mode != ENUMERATEDPATH)
  {	
    if (statespace_options.generate_mode & GRAPH)
      estream << "Random edge probability: "
              << statespace_options.edge_probability
              << '\n' + string(indent + 4, ' ');

    estream << "Propositional truth probability: "
	    << statespace_options.truth_probability
	    << "\n";
  }

  estream << "\n" + string(indent + 2, ' ') + "Random LTL formulae:\n"
             + string(indent + 4, ' ')
	  << formula_options.generation_parameters.length;

  if (formula_options.generation_parameters.max_length
        != formula_options.generation_parameters.length)
    estream << "..." << formula_options.generation_parameters.max_length;

  estream << string(" symbol")
             + (formula_options.generation_parameters.max_length == 1
		? "" : "s")
             + ", "
          << formula_options.generation_parameters.
               number_of_available_variables
          << string(" atomic proposition")
             + (formula_options.generation_parameters.
                  number_of_available_variables == 1 ? "" : "s")
             + '\n' + string(indent + 4, ' ');

  if (global_options.formula_change_interval == 0)
    estream << "Using a fixed LTL formula.\n" + string(indent + 4, ' ');
  else
  {
    estream << "New LTL formula will be generated after every ";
    if (global_options.formula_change_interval > 1)
    {
      estream << global_options.formula_change_interval;

      if (global_options.formula_change_interval < 10
	  || global_options.formula_change_interval > 20)
      {
	switch (global_options.formula_change_interval % 10)
        {
	  case 1 : estream << "st"; break;
	  case 2 : estream << "nd"; break;
	  case 3 : estream << "rd"; break;
	  default : estream << "th"; break;
	}
      }
      else
	estream << "th";

      estream << ' ';
    }

    estream << "round.\n" + string(indent + 4, ' ');
  }

  if (formula_options.generate_mode == NNF)
    estream << "Formulae will be generated in negation normal form.\n"
               + string(indent + 4, ' ');
  else if (formula_options.output_mode == NNF)
    estream << "Formulae will be converted into negation normal form.\n"
               + string(indent + 4, ' ');

  estream << "Operands in use (priority):\n" + string(indent + 6, ' ');

  bool first_printed = false;

  typedef pair<int, int> IntegerPair;

  for (vector<IntegerPair, ALLOC(IntegerPair) >::const_iterator op
	 = formula_options.generation_parameters.
             propositional_symbol_priorities.begin();
       op != formula_options.generation_parameters.
               propositional_symbol_priorities.end();
       ++op)
  {
    if (op->first == RandomFormulaParameters::NO_SYMBOL)
      continue;

    if (first_printed)
      estream << "; ";

    first_printed = true;

    switch (op->first)
    {
      case RandomFormulaParameters::ATOM :
	estream << "propositions";
	break;
      case RandomFormulaParameters::TRUE :
	estream << "true";
	break;
      case RandomFormulaParameters::FALSE :
	estream << "false";
	break;
      default :
	break;
    }

    estream << " (" << op->second << ')';
  }

  estream << '\n'
	  << string(indent + 4, ' ') + "Operators in use (priority):\n"
             + string(indent + 6, ' ');

  int number_of_printed_operators = 0;

  for (vector<IntegerPair, ALLOC(IntegerPair) >::const_iterator op
	 = formula_options.generation_parameters.
             long_formula_operators.begin();
       op != formula_options.generation_parameters.
               long_formula_operators.end();
       ++op)
  {
    if (op->first == RandomFormulaParameters::NO_SYMBOL)
      continue;

    if (number_of_printed_operators > 0)
    {
      estream << "; ";
      if (number_of_printed_operators % 6 == 0)
	estream << '\n' << string(indent + 6, ' ');
    }

    number_of_printed_operators++;

    switch (op->first)
    {
      case RandomFormulaParameters::AND :
	estream << "/\\";
	break;
      case RandomFormulaParameters::OR :
	estream << "\\/";
	break;
      case RandomFormulaParameters::IMPLICATION :
	estream << "->";
	break;
      case RandomFormulaParameters::EQUIVALENCE :
	estream << "<->";
	break;
      case RandomFormulaParameters::UNTIL :
	estream << 'U';
	break;
      case RandomFormulaParameters::V :
	estream << 'V';
	break;
      case RandomFormulaParameters::NEXT :
	estream << 'X';
	break;
      case RandomFormulaParameters::NOT :
	estream << '!';
	break;
      case RandomFormulaParameters::FINALLY :
	estream << "<>";
	break;
      case RandomFormulaParameters::GLOBALLY :
	estream << "[]";
	break;
      default :
	break;
    }

    estream << " (" << op->second;

    if (op->first == RandomFormulaParameters::NOT
	&& formula_options.generate_mode == NNF)
      estream << ", only preceding atoms";

    estream << ')';
  }

  estream << endl << endl;
}

/* ========================================================================= */
void Configuration::reset()
/* ----------------------------------------------------------------------------
 *
 * Description:   Resets the program configuration to the default
 *                configuration.
 *
 * Arguments:     None.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  global_options.verbosity = 5;
  global_options.interactive = ALWAYS;
  global_options.number_of_rounds = 10;
  global_options.statespace_change_interval = 10;
  global_options.formula_change_interval = 1;
  global_options.product_mode = GLOBAL;

  statespace_options.size = 20;
  statespace_options.atoms_per_state = 5;
  statespace_options.edge_probability = 0.2;
  statespace_options.truth_probability = 0.5;
  statespace_options.generate_mode = RANDOMCONNECTEDGRAPH;

  formula_options.allow_abbreviated_operators = true;
  formula_options.output_mode = NORMAL;
  formula_options.generate_mode = NORMAL;

  formula_options.generation_parameters.reset();
}
