/*
 *  Copyright (C) 1999, 2000  Heikki Tauriainen <Heikki.Tauriainen@hut.fi>
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

%{

#include <strstream>
#include "Configuration.h"



Configuration configuration;                        /* Global configuration. */



/******************************************************************************
 *
 * Prototypes for functions used for parsing the configuration file.
 *
 *****************************************************************************/

static void yyerror(char* error_message);           /* Displays an error
						     * message. (See the end of
						     * this file for the
						     * function definition.)
						     */

int yylex();                                        /* Reads the next token
						     * from the input (this
						     * function is provided by
						     * the lexer).
						     */

int getLineNumber();                                /* Returns the current
						     * input line number.
						     * See `configuration.lex'
						     * for the definition for
						     * this function.
						     */



/******************************************************************************
 *
 * Variables for parsing the configuration file.
 *
 *****************************************************************************/

static Configuration::AlgorithmInformation          /* Stores all the      */
  algorithm_information;                            /* information in a
						     * single `Algorithm'
						     * block in the
						     * configuration file.
						     */

static int algorithm_begin_line;                    /* Input file line number
						     * denoting the beginning
						     * of the most recently
						     * encountered `Algorithm'
						     * block.
						     */

static unsigned long int number_of_algorithms;      /* Used for counting the
						     * number of `Algorithm'
						     * blocks in the
						     * configuration file.
						     */

static int expected_token;                          /* Type of a token to be
						     * encountered next when
						     * reading the
						     * configuration file.
						     */

static int default_operator_priority;               /* Default priority for
						     * all LTL formula symbols.
						     */

static map<int, int, less<int>, ALLOC(int) >        /* Priorities for LTL */
  symbol_priority;                                  /* formula symbols.   */

static int current_block_type;                      /* Type of the current
						     * configuration block.
						     */

static int current_option_type;                     /* Type of the current
						     * option name.
						     */



/******************************************************************************
 *
 * Functions for performing various bound checks for the values of different
 * options in the configuration file.
 *
 *****************************************************************************/

/* ========================================================================= */
static void checkPriority(long int value)
/* ----------------------------------------------------------------------------
 *
 * Description:   Checks whether the priority specified in the configuration
 *                file for some formula symbol is between 0 and 10000.
 *                
 * Argument:      value  --  The priority assigned to the symbol in the
 *                           configuration file.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  if (value < 0 || value > 10000)
  {
    expected_token = UNKNOWN;
    yyerror("Priority value must be between 0 and 10000 inclusive.");
  }
}

/* ========================================================================= */
static void checkNumberOfPropositions(long int value)
/* ----------------------------------------------------------------------------
 *
 * Description:   Checks whether the number of propositions to use
 *                is nonnegative.
 *
 * Argument:      value  --  Number of propositions to use.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  if (value < 0)
  {
    expected_token = UNKNOWN;
    yyerror("Number of propositions must be nonnegative.");
  }
}

/* ========================================================================= */
static void checkProbability(double value)
/* ----------------------------------------------------------------------------
 *
 * Description:   Checks whether a probability value specified in the
 *                configuration file is between 0.0 and 1.0 inclusive.
 *
 * Argument:      value  --  A value supposed to denote a probability.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  if (value < 0.0 || value > 1.0)
  {
    expected_token = UNKNOWN;
    yyerror("Probability value must be between 0.0 and 1.0 inclusive.");
  }
}

%}


/******************************************************************************
 *
 * Declarations for terminal and nonterminal symbols used in the grammar rules
 * below.
 *
 *****************************************************************************/

/* Data types for configuration file option values. */

%union
{
  bool truth_value;

  long int integer;

  Configuration::InteractionMode interactivity_value;

  Configuration::FormulaMode formula_mode_value;

  Configuration::StateSpaceMode statespace_mode_value;

  Configuration::ProductMode product_type_value;

  double real;

  string *str;

  struct
  {
    long int min;
    long int max;
  } integer_interval;
}

/* Keywords. */

%token ALGORITHM NAME PROGRAMPATH PARAMETERS ENABLED

%token GLOBALOPTIONS VERBOSITY INTERACTIVE ROUNDS STATESPACECHANGEINTERVAL
       FORMULACHANGEINTERVAL RANDOMSEED SYNCHRONOUSPRODUCT

%token STATESPACEOPTIONS SIZE PROPOSITIONS EDGEPROBABILITY
       TRUTHPROBABILITY

%token FORMULAOPTIONS ABBREVIATEDOPERATORS OUTPUTMODE GENERATEMODE
       LENGTH PROPOSITIONPRIORITY TRUEPRIORITY FALSEPRIORITY
       DEFAULTOPERATORPRIORITY ANDPRIORITY ORPRIORITY IMPLICATIONPRIORITY
       EQUIVALENCEPRIORITY UNTILPRIORITY VPRIORITY NOTPRIORITY NEXTPRIORITY
       FINALLYPRIORITY GLOBALLYPRIORITY

/* Boolean constants. */

%token <truth_value> TRUTH_VALUE

/* Interactivity mode constants. */

%token <interactivity_value> INTERACTIVITY_VALUE

/* Formula output/generation mode constants. */

%token <formula_mode_value> FORMULA_MODE_VALUE

/* Statespace generation mode constants. */

%token <statespace_mode_value> STATESPACE_MODE_VALUE

/* Constants controlling the product space computation. */

%token <product_type_value> PRODUCT_TYPE_VALUE

/* Numbers. */

%token <integer> INTEGER
%token <real> REAL

/* Intervals of integers. */

%token <integer_interval> INTEGER_INTERVAL

/* String constants. */

%token <str> STRING_CONSTANT

/* Punctuation symbols. */

%token LBRACE RBRACE EQUALS
%token BLOCK_ID, OPTION_ID

/* The `unknown' token. */

%token UNKNOWN



/******************************************************************************
 *
 * Grammar rule definitions.
 *
 *****************************************************************************/

%%

configuration_file:          {
                               number_of_algorithms = 0;
                             }
                           configuration_blocks
                         ;

configuration_blocks:      /* empty */
                         | configuration_blocks
                             {
                               expected_token = BLOCK_ID;
                             }
                           configuration_block
                         ;

configuration_block:       algorithm_option_block
                         | global_option_block
                         | statespace_option_block
                         | formula_option_block
                         ;

algorithm_option_block:    ALGORITHM
                             {
			       current_block_type = ALGORITHM;

                               algorithm_begin_line = getLineNumber();

                               algorithm_information.name = 0;
                               algorithm_information.path_to_program = 0;
                               algorithm_information.extra_parameters = 0;
			       algorithm_information.enabled = true;

                               number_of_algorithms++;

                               expected_token = LBRACE;
                             }
                           LBRACE algorithm_options RBRACE
                             {
                               if (algorithm_information.name == 0)
                               {
                                 algorithm_information.name =
                                   new string("Algorithm #");
                                 ostrstream number_stream;
                                 number_stream << number_of_algorithms - 1
                                               << ends;
                                 number_stream.freeze();
                                 algorithm_information.name->append
                                   (number_stream.str());
                               }
                               configuration.algorithms.push_back
                                 (algorithm_information);
                               if (algorithm_information.path_to_program == 0)
                               {
                                 ostrstream msg;

                                 msg << "Missing program path information "
                                        "for '" + *(algorithm_information.name)
                                        + "' (lines "
				     << algorithm_begin_line
                                     << "--"
				     << getLineNumber()
				     << ")."
				     << ends;
                                 msg.freeze();
                                 throw Configuration::ConfigurationException
                                         (-1, msg.str());
                               }
                             }
                         ;

algorithm_options:         /* empty */
                         | algorithm_options
                             {
                               expected_token = OPTION_ID;
                             }
                           algorithm_option
                         ;

algorithm_option:          NAME
                             {
			       current_option_type = NAME;
                               expected_token = EQUALS;
                             }
                           EQUALS
                             {
                               expected_token = STRING_CONSTANT;
                             }
                           STRING_CONSTANT
                             {
                               algorithm_information.name = $5;
                             }

                         | PROGRAMPATH
                             {
			       current_option_type = PROGRAMPATH;
                               expected_token = EQUALS;
                             }
                           EQUALS
                             {
                               expected_token = STRING_CONSTANT;
                             }
                           STRING_CONSTANT
                             {
                               algorithm_information.path_to_program = $5;
                             }

                         | PARAMETERS
                             {
			       current_option_type = PARAMETERS;
                               expected_token = EQUALS;
                             }
                           EQUALS
                             {
                               expected_token = STRING_CONSTANT;
                             }
                           STRING_CONSTANT
                             {
                               algorithm_information.extra_parameters = $5;
                             }

                         | ENABLED
                             {
			       current_option_type = ENABLED;
			       expected_token = EQUALS;
			     }
                           EQUALS
                             {
			       expected_token = TRUTH_VALUE;
			     }
                           TRUTH_VALUE
                             {
			       algorithm_information.enabled = $5;
			     }
                         ;

global_option_block:       GLOBALOPTIONS
                             {
			       current_block_type = GLOBALOPTIONS;
                               expected_token = LBRACE;
                             }
                           LBRACE global_options RBRACE
                         ;

global_options:            /* empty */
                         | global_options
                             {
                               expected_token = OPTION_ID;
                             }
                           global_option
                         ;

global_option:             VERBOSITY
                             {
			       current_option_type = VERBOSITY;
                               expected_token = EQUALS;
                             }
                           EQUALS
                             {
                               expected_token = INTEGER;
                             }
                           INTEGER
                             {
                               if ($5 < 0 || $5 > 5)
                                 throw Configuration::ConfigurationException
                                         (getLineNumber(),
                                          "Verbosity must be between 0 and "
					  "5 inclusive.");
                               configuration.global_options.verbosity = $5;
                             }

                         | INTERACTIVE
                             {
			       current_option_type = INTERACTIVE;
                               expected_token = EQUALS;
                             }
                           EQUALS
                             {
                               expected_token = INTERACTIVITY_VALUE;
                             }
                           INTERACTIVITY_VALUE
                             {
                               configuration.global_options.interactive = $5;
                             }

                         | ROUNDS
                             {
			       current_option_type = ROUNDS;
                               expected_token = EQUALS;
                             }
                           EQUALS
                             {
                               expected_token = INTEGER;
                             }
                           INTEGER
                             {
                               if ($5 <= 0)
                                 throw Configuration::ConfigurationException
                                         (getLineNumber(),
					  "Number of rounds must be "
					  "positive.");
                               configuration.global_options.number_of_rounds =
                                 $5;
                             }

                         | STATESPACECHANGEINTERVAL
                             {
			       current_option_type = STATESPACECHANGEINTERVAL;
                               expected_token = EQUALS;
                             }
                           EQUALS
                             {
                               expected_token = INTEGER;
                             }
                           INTEGER
                             {
                               if ($5 < 0)
                                 throw Configuration::ConfigurationException
                                         (getLineNumber(),
					  "Length of interval must be "
                                          "nonnegative.");
                               configuration.global_options.
                                 statespace_change_interval = $5;
                             }

                         | FORMULACHANGEINTERVAL
                             {
			       current_option_type = FORMULACHANGEINTERVAL;
                               expected_token = EQUALS;
                             }
                           EQUALS
                             {
                               expected_token = INTEGER;
                             }
                           INTEGER
                             {
                               if ($5 < 0)
                                 throw Configuration::ConfigurationException
                                         (getLineNumber(),
                                          "Length of interval must be "
                                          "nonnegative.");
                               configuration.global_options.
                                 formula_change_interval = $5;
                             }

                         | RANDOMSEED
                             {
			       current_option_type = RANDOMSEED;
                               expected_token = EQUALS;
                             }
                           EQUALS
                             {
                               expected_token = INTEGER;
                             }
                           INTEGER
                             {
			       SRAND($5);
                             }

                         | SYNCHRONOUSPRODUCT
                             {
			       current_option_type = SYNCHRONOUSPRODUCT;
			       expected_token = EQUALS;
			     }
                           EQUALS
                             {
			       expected_token = PRODUCT_TYPE_VALUE;
			     }
                           PRODUCT_TYPE_VALUE
                             {
			       configuration.global_options.product_mode =
				 $5;
			     }
                         ;

statespace_option_block:   STATESPACEOPTIONS
                             {
			       current_block_type = STATESPACEOPTIONS;
                               expected_token = LBRACE;
                             }
                           LBRACE statespace_options RBRACE
                         ;

statespace_options:        /* empty */
                         | statespace_options
                             {
                               expected_token = OPTION_ID;
                             }
                           statespace_option
                         ;

statespace_option:         SIZE
                             {
			       current_option_type = SIZE;
                               expected_token = EQUALS;
                             }
                           EQUALS
                             {
                               expected_token = INTEGER_INTERVAL;
                             }
                           statespace_size

                         | PROPOSITIONS
                             {
			       current_option_type = PROPOSITIONS;
                               expected_token = EQUALS;
                             }
                           EQUALS
                             {
                               expected_token = INTEGER;
                             }
                           INTEGER
                             {
                               checkNumberOfPropositions($5);
                               configuration.statespace_options.
                                 atoms_per_state = $5;
                             }

                         | EDGEPROBABILITY
                             {
			       current_option_type = EDGEPROBABILITY;
                               expected_token = EQUALS;
                             }
                           EQUALS
                             {
                               expected_token = REAL;
                             }
                           REAL
                             {
                               checkProbability($5);
                               configuration.statespace_options.
                                 edge_probability = $5;
                             }

                         | TRUTHPROBABILITY
                             {
			       current_option_type = TRUTHPROBABILITY;
                               expected_token = EQUALS;
                             }
                           EQUALS
                             {
                               expected_token = REAL;
                             }
                           REAL
                             {
                               checkProbability($5);
                               configuration.statespace_options.
                                 truth_probability = $5;
                             }

                         | GENERATEMODE
                             {
			       current_option_type = GENERATEMODE;
			       expected_token = EQUALS;
			     }
                           EQUALS
                             {
			       expected_token = STATESPACE_MODE_VALUE;
			     }
                           STATESPACE_MODE_VALUE
                             {
			       configuration.statespace_options.
				 generate_mode = $5;
			     }
                         ;

formula_option_block:      FORMULAOPTIONS
                             {
			       current_block_type = FORMULAOPTIONS;
                               expected_token = LBRACE;
                             }
                           LBRACE formula_options RBRACE
                         ;

formula_options:           /* empty */
                         | formula_options
                             {
                               expected_token = OPTION_ID;
                             }
                           formula_option
                         ;

formula_option:            ABBREVIATEDOPERATORS
                             {
			       current_option_type = ABBREVIATEDOPERATORS;
                               expected_token = EQUALS;
                             }
                           EQUALS
                             {
                               expected_token = TRUTH_VALUE;
                             }
                           TRUTH_VALUE
                             {
                               configuration.formula_options.
                                 allow_abbreviated_operators = $5;
                             }

                         | OUTPUTMODE
                             {
			       current_option_type = OUTPUTMODE;
                               expected_token = EQUALS;
                             }
                           EQUALS
                             {
                               expected_token = FORMULA_MODE_VALUE;
                             }
                           FORMULA_MODE_VALUE
                             {
                               configuration.formula_options.
                                 output_mode = $5;
                             }

                         | GENERATEMODE
                             {
			       current_option_type = GENERATEMODE;
                               expected_token = EQUALS;
                             }
                           EQUALS
                             {
                               expected_token = FORMULA_MODE_VALUE;
                             }
                           FORMULA_MODE_VALUE
                             {
                               configuration.formula_options.
                                 generate_mode = $5;
                             }

                         | PROPOSITIONS
                             {
			       current_option_type = PROPOSITIONS;
                               expected_token = EQUALS;
                             }
                           EQUALS
                             {
                               expected_token = INTEGER;
                             }
                           INTEGER
                             {
                               checkNumberOfPropositions($5);
                               configuration.formula_options.
                                 generation_parameters.
                                 number_of_available_variables = $5;
                             }

                         | LENGTH
                             {
			       current_option_type = LENGTH;
                               expected_token = EQUALS;
                             }
                           EQUALS
                             {
                               expected_token = INTEGER_INTERVAL;
                             }
                           formula_length

                         | PROPOSITIONPRIORITY
                             {
			       current_option_type = PROPOSITIONPRIORITY;
                               expected_token = EQUALS;
                             }
                           EQUALS
                             {
                               expected_token = INTEGER;
                             }
                           INTEGER
                             {
                               checkPriority($5);
                               symbol_priority[RandomFormulaParameters::ATOM] =
                                 $5;
                             }

                         | TRUEPRIORITY
                             {
			       current_option_type = TRUEPRIORITY;
                               expected_token = EQUALS;
                             }
                           EQUALS
                             {
                               expected_token = INTEGER;
                             }
                           INTEGER
                             {
                               checkPriority($5);
                               symbol_priority[RandomFormulaParameters::TRUE] =
                                 $5;
                             }

                         | FALSEPRIORITY
                             {
			       current_option_type = FALSEPRIORITY;
                               expected_token = EQUALS;
                             }
                           EQUALS
                             {
                               expected_token = INTEGER;
                             }
                           INTEGER
                             {
                               checkPriority($5);
                               symbol_priority[RandomFormulaParameters::FALSE]
                                 = $5;
                             }

                         | DEFAULTOPERATORPRIORITY
                             {
			       current_option_type = DEFAULTOPERATORPRIORITY;
                               expected_token = EQUALS;
                             }
                           EQUALS
                             {
                               expected_token = INTEGER;
                             }
                           INTEGER
                             {
                               if ($5 <= 0 || $5 >= 10000)
				 throw Configuration::ConfigurationException
				         (getLineNumber(),
					  "Default operator priority must be "
					  "between 1 and 10000 inclusive.");
                               default_operator_priority = $5;
                             }

                         | ANDPRIORITY
                             {
			       current_option_type = ANDPRIORITY;
                               expected_token = EQUALS;
                             }
                           EQUALS
                             {
                               expected_token = INTEGER;
                             }
                           INTEGER
                             {
                               checkPriority($5);
                               symbol_priority[RandomFormulaParameters::AND]
				 = $5;
                             }

                         | ORPRIORITY
                             {
			       current_option_type = ORPRIORITY;
                               expected_token = EQUALS;
                             }
                           EQUALS
                             {
                               expected_token = INTEGER;
                             }
                           INTEGER
                             {
                               checkPriority($5);
                               symbol_priority[RandomFormulaParameters::OR]
                                 = $5;
                             }

                         | IMPLICATIONPRIORITY
                             {
			       current_option_type = IMPLICATIONPRIORITY;
                               expected_token = EQUALS;
                             }
                           EQUALS
                             {
                               expected_token = INTEGER;
                             }
                           INTEGER
                             {
                               checkPriority($5);
                               symbol_priority[RandomFormulaParameters::
                                               IMPLICATION] = $5;
                             }

                         | EQUIVALENCEPRIORITY
                             {
			       current_option_type = EQUIVALENCEPRIORITY;
                               expected_token = EQUALS;
                             }
                           EQUALS
                             {
                               expected_token = INTEGER;
                             }
                           INTEGER
                             {
                               checkPriority($5);
                               symbol_priority[RandomFormulaParameters::
                                               EQUIVALENCE] = $5;
                             }

                         | UNTILPRIORITY
                             {
			       current_option_type = UNTILPRIORITY;
                               expected_token = EQUALS;
                             }
                           EQUALS
                             {
                               expected_token = INTEGER;
                             }
                           INTEGER
                             {
                               checkPriority($5);
                               symbol_priority[RandomFormulaParameters::UNTIL]
                                 = $5;
                             }

                         | VPRIORITY
                             {
			       current_option_type = VPRIORITY;
                               expected_token = EQUALS;
                             }
                           EQUALS
                             {
                               expected_token = INTEGER;
                             }
                           INTEGER
                             {
                               checkPriority($5);
                               symbol_priority[RandomFormulaParameters::V]
                                 = $5;
                             }

                         | NOTPRIORITY
                             {
			       current_option_type = NOTPRIORITY;
                               expected_token = EQUALS;
                             }
                           EQUALS
                             {
                               expected_token = INTEGER;
                             }
                           INTEGER
                             {
                               checkPriority($5);
                               symbol_priority[RandomFormulaParameters::NOT]
                                 = $5;
                             }

                         | NEXTPRIORITY
                             {
			       current_option_type = NEXTPRIORITY;
                               expected_token = EQUALS;
                             }
                           EQUALS
                             {
                               expected_token = INTEGER;
                             }
                           INTEGER
                             {
                               checkPriority($5);
                               symbol_priority[RandomFormulaParameters::NEXT]
                                 = $5;
                             }

                         | FINALLYPRIORITY
                             {
			       current_option_type = FINALLYPRIORITY;
                               expected_token = EQUALS;
                             }
                           EQUALS
                             {
                               expected_token = INTEGER;
                             }
                           INTEGER
                             {
                               checkPriority($5);
                               symbol_priority[RandomFormulaParameters::
                                               FINALLY] = $5;
                             }

                         | GLOBALLYPRIORITY
                             {
			       current_option_type = GLOBALLYPRIORITY;
                               expected_token = EQUALS;
                             }
                           EQUALS
                             {
                               expected_token = INTEGER;
                             }
                           INTEGER
                             {
                               checkPriority($5);
                               symbol_priority[RandomFormulaParameters::
                                               GLOBALLY] = $5;
                             }
                         ;

formula_length:            INTEGER
                             {
                               if ($1 <= 0)
                                 throw Configuration::ConfigurationException
                                         (getLineNumber(),
				          "Formula length must be positive.");
                               configuration.formula_options.
                                 generation_parameters.length = $1;
			       configuration.formula_options.
				 generation_parameters.max_length = $1;
                             }

                         | INTEGER_INTERVAL
                             {
			       if ($1.min <= 0)
				 throw Configuration::ConfigurationException
				         (getLineNumber(),
					  "Minimum formula length must be"
					  " positive.");
			       
			       if ($1.max < $1.min)
				 throw Configuration::ConfigurationException
				         (getLineNumber(),
					  "Minimum formula length exceeds "
                                          "the maximum formula length.");

			       configuration.formula_options.
				 generation_parameters.length = $1.min;
			       configuration.formula_options.
				 generation_parameters.max_length = $1.max;
			     }
                         ;

statespace_size:           INTEGER
                             {
                               if ($1 <= 0)
                                 throw Configuration::ConfigurationException
                                         (getLineNumber(),
					  "Size of state space must be "
					  "positive.");
                               configuration.statespace_options.size = $1;
			       configuration.statespace_options.max_size = $1;
                             }

                         | INTEGER_INTERVAL
                             {
			       if ($1.min <= 0)
				 throw Configuration::ConfigurationException
				         (getLineNumber(),
					  "Minimum state space size must be "
					  "positive.");

			       if ($1.max < $1.min)
				 throw Configuration::ConfigurationException
				         (getLineNumber(),
					  "Minimum state space size exceeds "
					  "the maximum state space size.");

			       configuration.statespace_options.size = $1.min;
			       configuration.statespace_options.max_size
				 = $1.max;
			     }
                         ;

%%



#include "Configuration.yy.cc"



/******************************************************************************
 *
 * Function for displaying error messages.
 *
 *****************************************************************************/

void yyerror(char* error_message)
{
  string msg;

  switch (expected_token)
  {
    case BLOCK_ID :
      msg = string("Unknown block identifier ('") + yytext + "').";
      break;
    case OPTION_ID :
      msg = string("Unknown option identifier ('") + yytext + "').";
      break;
    case LBRACE :
      msg = "'{' expected after block name.";
      break;
    case EQUALS :
      msg = "'=' expected after option name.";
      break;
    case TRUTH_VALUE :
      msg = "Truth value expected as option value.";
      break;
    case INTERACTIVITY_VALUE :
      msg = "Interactivity mode expected as option value.";
      break;
    case FORMULA_MODE_VALUE :
      msg = "Formula generation mode expected as option value.";
      break;
    case STATESPACE_MODE_VALUE :
      msg = "Statespace generation mode expected as option value.";
      break;
    case PRODUCT_TYPE_VALUE :
      msg = "Product computation mode expected as option value.";
      break;
    case INTEGER :
      msg = "Nonnegative integer expected as option value.";
      break;
    case INTEGER_INTERVAL :
      msg = "Nonnegative integer or an integer interval expected as option"
	    " value.";
      break;
    case REAL :
      msg = "Nonnegative real number expected as option value.";
      break;
    case STRING_CONSTANT :
      msg = "String constant expected as option value.";
      break;
    default :
      msg = error_message;
      break;
  }

  throw Configuration::ConfigurationException(getLineNumber(), msg);
}
