/*
 *  Copyright (C) 1999, 2000  Heikki Tauriainen <Heikki.Tauriainen@hut.fi>
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include "FormulaString.h"

namespace FormulaString
{

/******************************************************************************
 *
 * Function definitions for class FormulaString.
 *
 *****************************************************************************/

/* ========================================================================= */
unsigned long int FormulaString::randomize
  (const RandomFormulaParameters& randomization_parameters)
/* ----------------------------------------------------------------------------
 *
 * Description:   Initializes a FormulaString with a random formula, whose
 *                generation is parameterized by the information in the
 *                function argument.
 *
 * Argument:      randomization_parameters  --  Information controlling the
 *                                              generation of the formula
 *                                              (length of the formula, maximum
 *                                              number of different variables 
 *                                              in the formula, priorities for
 *                                              different operators).
 *
 * Returns:       Length of the random formula.
 *
 * ------------------------------------------------------------------------- */
{
  ostrstream formula_stream;

  unsigned long int length = randomization_parameters.length;

  if (randomization_parameters.max_length > length)
    length += LRAND % (randomization_parameters.max_length + 1 - length);

  recGenFormula(formula_stream, length, randomization_parameters);

  formula_stream << ends;
  formula_stream.freeze();

  formula = formula_stream.str();

  return length;
}

/* ========================================================================= */
void FormulaString::read(istream& input_stream)
/* ----------------------------------------------------------------------------
 *
 * Description:   Initializes a FormulaString by reading a formula from a
 *                stream.
 *
 * Argument:      input_stream  --  A reference to an input stream.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  formula = "";

  TraversalInfo<string*> info;
  info.data = &formula;

  FormulaString::template
    traverseStream<ReadAtom, ReadSymbol<LTL_TRUE>, ReadSymbol<LTL_FALSE>, 
                   ReadSymbol<LTL_NOT>, ReadSymbol<LTL_NEXT>,
                   ReadSymbol<LTL_FINALLY>, ReadSymbol<LTL_GLOBALLY>,
                   ReadSymbol<LTL_AND>, ReadSymbol<LTL_OR>,
                   ReadSymbol<LTL_IMPLICATION>, ReadSymbol<LTL_EQUIVALENCE>,
                   ReadSymbol<LTL_UNTIL>, ReadSymbol<LTL_V> >
    (PREORDER, input_stream, info);

  formula.resize(formula.size() - 1);
}

/* ========================================================================= */
void FormulaString::read(const char* filename)
/* ----------------------------------------------------------------------------
 *
 * Description:   Initializes a FormulaString by reading a formula from a file.
 *
 * Argument:      filename  --  Name of the file.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  formula = "";

  ifstream formula_file;
  formula_file.open(filename, ios::in);

  if (!formula_file.good())
    throw FileOpenException(filename);

  try
  {
    read(formula_file);
    formula_file.close();
  }
  catch (...)
  {
    formula_file.close();
    throw;
  }
}

/*
 *  Strings to be used for instatiating the `traverse' template in function
 *  FormulaString::prettyPrint.
 */

char LTL_PRETTYPRINT_TRUE[]        = "true";
char LTL_PRETTYPRINT_FALSE[]       = "false";
char LTL_PRETTYPRINT_NOT[]         = "!";
char LTL_PRETTYPRINT_NEXT[]        = "X";
char LTL_PRETTYPRINT_FINALLY[]     = "<>";
char LTL_PRETTYPRINT_GLOBALLY[]    = "[]";
char LTL_PRETTYPRINT_AND[]         = "/\\";
char LTL_PRETTYPRINT_OR[]          = "\\/";
char LTL_PRETTYPRINT_IMPLICATION[] = "->";
char LTL_PRETTYPRINT_EQUIVALENCE[] = "<->";
char LTL_PRETTYPRINT_UNTIL[]       = "U";
char LTL_PRETTYPRINT_V[]           = "V";

/* ========================================================================= */
void FormulaString::prettyPrint(ostream& output_stream = cout)
/* ----------------------------------------------------------------------------
 *
 * Description:   Writes a FormulaString in a parenthesized infix notation to a
 *                stream.
 *
 * Argument:      output_stream  --  Reference to the output stream.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  TraversalInfo<Exceptional_ostream*> info;
  Exceptional_ostream eoutput_stream(&output_stream,
                                     ios::failbit | ios::badbit);
  info.data = &eoutput_stream;

  traverse<FormatAtom,
           FormatConstant<LTL_PRETTYPRINT_TRUE>,
           FormatConstant<LTL_PRETTYPRINT_FALSE>,
           FormatUnary<LTL_PRETTYPRINT_NOT>,
           FormatUnary<LTL_PRETTYPRINT_NEXT>,
           FormatUnary<LTL_PRETTYPRINT_FINALLY>,
           FormatUnary<LTL_PRETTYPRINT_GLOBALLY>,
           FormatBinary<LTL_PRETTYPRINT_AND>,
           FormatBinary<LTL_PRETTYPRINT_OR>,
           FormatBinary<LTL_PRETTYPRINT_IMPLICATION>,
           FormatBinary<LTL_PRETTYPRINT_EQUIVALENCE>,
           FormatBinary<LTL_PRETTYPRINT_UNTIL>,
           FormatBinary<LTL_PRETTYPRINT_V> >
    (PREORDER | INORDER | POSTORDER, info);
}

/* ========================================================================= */
void FormulaString::negate()
/* ----------------------------------------------------------------------------
 *
 * Description:   Negates a FormulaString.
 *
 * Arguments:     None.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  string negated_formula = "! ";
  negated_formula += formula;
  formula = negated_formula;
}

/* ========================================================================= */
void FormulaString::toNegationNormalForm()
/* ----------------------------------------------------------------------------
 *
 * Description:   Converts the FormulaString into negation normal form, i.e.
 *                to a representation having no `abbreviated' operators
 *                (`finally', `globally', implication and equivalence) with
 *                all negations preceding propositional variables.
 *
 * Arguments:     None.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  TraversalInfo<NNFConvertData> info;
  info.data.negate_stack.push(false);
  info.data.result = new string;

  traverse<AtomToNNF,
           ConstantToNNF<true>,
           ConstantToNNF<false>,
           NegationToNNF,
           NextToNNF,
           AbbreviatedUnaryToNNF<LTL_UNTIL, LTL_V, false>,
           AbbreviatedUnaryToNNF<LTL_V, LTL_UNTIL, true>,
           BinaryToNNF<LTL_AND, LTL_OR, true, true>,
           BinaryToNNF<LTL_OR, LTL_AND, true, true>,
           BinaryToNNF<LTL_OR, LTL_AND, false, true>,
           EquivalenceToNNF,
           BinaryToNNF<LTL_UNTIL, LTL_V, true, true>,
           BinaryToNNF<LTL_V, LTL_UNTIL, true, true> >
    (PREORDER | INORDER | POSTORDER, info);

  formula = *info.data.result;
  delete info.data.result;
}

/* ========================================================================= */
FormulaString::FormulaSymbolType FormulaString::parseSymbol(char symbol)
/* ----------------------------------------------------------------------------
 *
 * Description:   Tells the type of a formula symbol.
 *    
 * Argument:      symbol  --  The symbol.
 *
 * Returns:       The type of the symbol (PROPOSITIONALSYMBOL, UNARYOPERATOR,
 *                BINARYOPERATOR or UNKNOWN).
 *
 * ------------------------------------------------------------------------- */
{
  switch (symbol)
  {
    case LTL_PROPOSITION :
    case LTL_TRUE :
    case LTL_FALSE :
      return PROPOSITIONALSYMBOL;
        
    case LTL_NOT :
    case LTL_NEXT :
    case LTL_FINALLY :
    case LTL_GLOBALLY :
      return UNARYOPERATOR;
        
    case LTL_AND :
    case LTL_OR :
    case LTL_IMPLICATION :
    case LTL_EQUIVALENCE :
    case LTL_UNTIL :
    case LTL_V :
      return BINARYOPERATOR;
        
    default :
      return UNKNOWN;
  }
}

/* ========================================================================= */
void FormulaString::recGenFormula
  (ostream& output_stream, unsigned long int length_of_formula,
   const RandomFormulaParameters& randomization_parameters)
/* ----------------------------------------------------------------------------
 *
 * Description:   Function for recursively generating a random LTL formula of
 *                a given length. The generated formula is written to a stream.
 *
 * Arguments:     output_stream             --  A reference to an output
 *                                              stream.
 *                length_of_formula         --  The length of the formula to be
 *                                              generated.
 *                randomization_parameters  --  A reference to a data structure
 *                                              with information controlling
 *                                              the generation of the formula
 *                                              (length of the formula, maximum
 *                                              number of different variables
 *                                              in the formula, priorities for
 *                                              different operators).
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  Exceptional_ostream eoutput_stream(&output_stream,
                                     ios::failbit | ios::badbit);

  typedef pair<int, int> SymbolPriority;

  vector<SymbolPriority, ALLOC(SymbolPriority) >::const_iterator
    symbol_priority;

  long int x = LRAND;

  /*
   *  Select a list of allowable symbols according to the target length.
   *  If the length is 1, only propositional variables and constants are
   *  allowed. If the length is 2, only unary operators are allowed.
   *  Otherwise all unary and binary operators are allowed.
   */

  switch (length_of_formula)
  {
    case 1 :
      x %= randomization_parameters.propositional_symbol_priorities[0].
             second;
      symbol_priority = randomization_parameters.
                          propositional_symbol_priorities.begin();
      break;
    case 2 :
      x %= randomization_parameters.short_formula_operators[0].second;
      symbol_priority = randomization_parameters.
                          short_formula_operators.begin();
      break;
    default :
      x %= randomization_parameters.long_formula_operators[0].second;
      symbol_priority = randomization_parameters.
                          long_formula_operators.begin();
      break;
  }

  /*
   *  Using the selected list, choose a symbol from the list randomly,
   *  according to the priority distribution of the different symbols in
   *  the list. The list consists of pairs <symbol, priority>; however,
   *  the first element of the list does not denote any symbol: instead,
   *  it gives the sum of the priorities of the symbols in the list.
   */

  ++x;
  ++symbol_priority;

  int sum = 0;
  do
  {
    sum += (*symbol_priority).second;
    if (sum < x)
      ++symbol_priority;
  }
  while (sum < x);

  switch ((*symbol_priority).first)
  {
    /*
     *  Propositional atom: Append a random propositional variable to the end
     *                      of the formula.
     */

    case RandomFormulaParameters::ATOM :
      eoutput_stream << LTL_PROPOSITION << (LRAND %
                                            randomization_parameters.
                                            number_of_available_variables);
      break;

    /*
     *  Propositional constants.
     */

    case RandomFormulaParameters::TRUE :
      eoutput_stream << LTL_TRUE;
      break;

    case RandomFormulaParameters::FALSE :
      eoutput_stream << LTL_FALSE;
      break;

    /*
     *  Unary operators: First append a random unary operator to the end of
     *                   the formula, then recursively generate a random
     *                   formula of length `length_of_formula' - 1.
     */

    case RandomFormulaParameters::NOT :
    case RandomFormulaParameters::NEXT :
    case RandomFormulaParameters::FINALLY :
    case RandomFormulaParameters::GLOBALLY :
      eoutput_stream <<
        (((*symbol_priority).first == RandomFormulaParameters::NOT ? LTL_NOT :
          (*symbol_priority).first == RandomFormulaParameters::NEXT ?
                                      LTL_NEXT :
          (*symbol_priority).first == RandomFormulaParameters::FINALLY ?
                                      LTL_FINALLY :
          LTL_GLOBALLY)) << ' ';
      recGenFormula(output_stream, length_of_formula - 1,
                    randomization_parameters);
      break;

    /*
     *  Binary operators: First append a random binary operator to the end of
     *                    the formula, then recursively generate two formulae
     *                    with their combined length `length_of_formula' - 1.
     */

    case RandomFormulaParameters::AND :
    case RandomFormulaParameters::OR :
    case RandomFormulaParameters::UNTIL :
    case RandomFormulaParameters::V :
    case RandomFormulaParameters::IMPLICATION :
    case RandomFormulaParameters::EQUIVALENCE :
      {
        eoutput_stream <<
          (((*symbol_priority).first == RandomFormulaParameters::AND ?
                                        LTL_AND :
            (*symbol_priority).first == RandomFormulaParameters::OR ? LTL_OR :
            (*symbol_priority).first == RandomFormulaParameters::UNTIL ?
                                        LTL_UNTIL :
            (*symbol_priority).first == RandomFormulaParameters::V ? LTL_V :
            (*symbol_priority).first == RandomFormulaParameters::IMPLICATION ?
            LTL_IMPLICATION : LTL_EQUIVALENCE)) << ' ';

	unsigned long int len = (LRAND % (length_of_formula - 2)) + 1;

        recGenFormula(output_stream, len, randomization_parameters);
        output_stream << ' ';
        recGenFormula(output_stream, length_of_formula - len - 1,
                      randomization_parameters);
      }
      break;
  }
}



/******************************************************************************
 *
 * Function definitions for class FormulaString::ReadAtom.
 *
 *****************************************************************************/

/* ========================================================================= */
inline void FormulaString::ReadAtom::eval
  (FormulaString::TraversalInfo<string*>& info, const string& atom)
/* ----------------------------------------------------------------------------
 *
 * Description:   Appends a propositional atom read from some stream to the
 *                string being constructed. Used by
 *                FormulaString::read(istream&) when traversing a formula in 
 *                an input stream.
 *
 * Arguments:     info  --  A reference to the TraversalInfo structure
 *                          containing the partially constructed result.
 *                atom  --  Propositional atom read from a stream.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  *(info.data) += string(atom) + ' ';
}



/******************************************************************************
 *
 * Function definitions for class FormulaString::ReadSymbol.
 *
 *****************************************************************************/

/* ========================================================================= */
template<char C>
inline void FormulaString::ReadSymbol::eval
  (FormulaString::TraversalInfo<string*>& info)
/* ----------------------------------------------------------------------------
 * 
 * Description:   Appends a symbol (a propositional constant or an operator,
 *                given as a template parameter) read from some stream to the
 *                string under construction. Used by
 *                FormulaString::read(istream&) when traversing a formula in
 *                an input stream.
 *
 * Arguments:     info  --  A reference to the TraversalInfo structure
 *                          containing the partially constructed result.
 *
 * Returns:       Nothing.
 * 
 * ------------------------------------------------------------------------- */
{
  *(info.data) += C;
  *(info.data) += ' ';
}



/******************************************************************************
 *
 * Function definitions for class FormulaString::FormatAtom.
 *
 *****************************************************************************/

/* ========================================================================= */
inline void FormulaString::FormatAtom::eval
  (FormulaString::TraversalInfo<Exceptional_ostream*>& info,
   const string& atom)
/* ----------------------------------------------------------------------------
 *
 * Description:   Writes a propositional variable to a stream. Used by
 *                FormulaString::prettyPrint() when traversing a formula to
 *                translate it into parenthesized infix notation.
 *
 * Arguments:     info  --  A reference to a TraversalInfo structure containing
 *                          a pointer to an output stream.
 *                atom  --  A propositional variable in string representation.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  *(info.data) << atom;
}



/******************************************************************************
 *
 * Function definitions for class FormulaString::FormatConstant.
 *
 *****************************************************************************/

/* ========================================================================= */
template<char* C>
inline void FormulaString::FormatConstant::eval
  (FormulaString::TraversalInfo<Exceptional_ostream*>& info)
/* ----------------------------------------------------------------------------
 *
 * Description:   Writes a propositional constant (given as a template
 *                parameter) to a stream. Used by FormulaString::prettyPrint()
 *                when traversing a FormulaString to translate it into
 *                parenthesized infix notation.
 *
 * Argument:      info  --  A reference to a TraversalInfo structure containing
 *                          a pointer to an output stream.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  *(info.data) << C;
}



/******************************************************************************
 *
 * Function definitions for class FormulaString::FormatUnary.
 *
 *****************************************************************************/

/* ========================================================================= */
template<char* C>
inline void FormulaString::FormatUnary::eval
  (FormulaString::TraversalInfo<Exceptional_ostream*>& info)
/* ----------------------------------------------------------------------------
 *
 * Description:   Writes an unary operator (given as a template parameter) to a
 *                stream. Used by FormulaString::prettyPrint() when traversing
 *                a FormulaString to translate it into parenthesized infix
 *                notation.
 *
 * Argument:      info  --  A reference to a TraversalInfo structure containing
 *                          a pointer to an output stream.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  if (info.operand == 0)
    *(info.data) << C << ' ';
}



/******************************************************************************
 *
 * Function definitions for class FormulaString::FormatBinary.
 *
 *****************************************************************************/

/* ========================================================================= */
template<char* C>
inline void FormulaString::FormatBinary::eval
  (FormulaString::TraversalInfo<Exceptional_ostream*>& info)
/* ----------------------------------------------------------------------------
 *
 * Description:   Translates a binary subformula (the binary operator is given
 *                as a template parameter) of a FormulaString object into
 *                parenthesized infix notation. Used by
 *                FormulaString::prettyPrint() when traversing a FormulaString.
 *
 * Argument:      info  --  A reference to a TraversalInfo structure containing
 *                          a pointer to an output stream.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  switch (info.operand)
  {
    case 0 : *(info.data) << '('; break;
    case 1 : *(info.data) << ' ' << C << ' '; break;
    default : *(info.data) << ')'; break;
  }
}



/******************************************************************************
 *
 * Function definitions for class FormulaString::AtomToNNF.
 *
 *****************************************************************************/

/* ========================================================================= */
inline void FormulaString::AtomToNNF::eval
  (FormulaString::TraversalInfo<NNFConvertData>& info, const string& atom)
/* ----------------------------------------------------------------------------
 *
 * Description:   Converts a propositional atom to negation normal form.
 *                Used by FormulaString::toNegationNormalForm() when traversing
 *                a FormulaString to perform the conversion.
 *
 *                This function implements the conversions
 *                    p  ==>  p  and
 *                   !p  ==> !p  .
 *
 * Arguments:     info  --  A reference to a TraversalInfo object containing
 *                          the information whether the propositional
 *                          variable needs to be negated, in addition with
 *                          the partially constructed result.
 *                atom  --  A propositional variable in string representation.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  if (info.data.negate_stack.top())
  {
    *(info.data.result) += LTL_NOT;
    *(info.data.result) += ' ';
  }
  
  *(info.data.result) += atom;
  
  info.data.negate_stack.pop();
  if (!info.data.negate_stack.empty())
    *(info.data.result) += ' ';
}



/******************************************************************************
 *
 * Function definitions for class FormulaString::ConstantToNNF.
 *
 *****************************************************************************/

/* ========================================================================= */
template<bool B>
inline void FormulaString::ConstantToNNF::eval
  (FormulaString::TraversalInfo<NNFConvertData>& info)
/* ----------------------------------------------------------------------------
 *
 * Description:   Converts a propositional constant to negation normal form.
 *                Used by FormulaString::toNegationNormalForm() when traversing
 *                a FormulaString to perform the conversion.
 *
 *                This function implements the conversions
 *                   true  ==>  true        false  ==>  false
 *                  !true  ==> false       !false  ==>  true
 *
 * Argument:      info  --  A reference to a TraversalInfo object containing
 *                          the information whether the propositional
 *                          constant needs to be negated, in addition with
 *                          the partially constructed result.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  *(info.data.result) += ((info.data.negate_stack.top() == B) ? LTL_FALSE :
                                                                LTL_TRUE);
  info.data.negate_stack.pop();
  if (!info.data.negate_stack.empty())
    *(info.data.result) += ' ';
}



/******************************************************************************
 *
 * Function definitions for class FormulaString::NegationToNNF.
 *
 *****************************************************************************/

/* ========================================================================= */
inline void FormulaString::NegationToNNF::eval
  (FormulaString::TraversalInfo<NNFConvertData>& info)
/* ----------------------------------------------------------------------------
 *
 * Description:   Converts a negated propositional formula to negation normal
 *                form. Used by FormulaString::toNegationNormalForm() when
 *                traversing a FormulaString to perform the conversion.
 *
 *                This function implements the conversions
 *                    !p   ==>  NNF(p)   and
 *                   !!p   ==>  p        .
 *
 * Argument:      info  --  A reference to a TraversalInfo object containing
 *                          the information whether a negation needs to be
 *                          pushed over the negation operator (thus cancelling
 *                          the negation), in addition with the partially
 *                          constructed result.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  if (info.operand == 0)
    info.data.negate_stack.top() = !info.data.negate_stack.top();
}


 
/******************************************************************************
 *
 * Function definitions for class FormulaString::NextToNNF.
 *
 *****************************************************************************/

/* ========================================================================= */
inline void FormulaString::NextToNNF::eval
  (FormulaString::TraversalInfo<NNFConvertData>& info)
/* ----------------------------------------------------------------------------
 *
 * Description:   Converts a propositional formula with a Next operator to
 *                negation normal form. Used by 
 *                FormulaString::toNegationNormalForm() when traversing a
 *                FormulaString to perform the conversion.
 *
 *                This function implements the conversions
 *                    Xp  ==>  Xp   and
 *                   !Xp  ==>  X!p  .
 *
 * Argument:      info  --  A reference to a TraversalInfo object containing
 *                          the information whether a negation needs to be
 *                          pushed over the Next operator, in addition with the
 *                          partially constructed result.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  if (info.operand == 0)
  {
    *(info.data.result) += LTL_NEXT;
    *(info.data.result) += ' ';
  }
}



/******************************************************************************
 *
 * Function definitions for class FormulaString::AbbreviatedUnaryToNNF.
 *
 *****************************************************************************/

/* ========================================================================= */
template<char OP1, char OP2, bool C>
inline void FormulaString::AbbreviatedUnaryToNNF::eval
  (FormulaString::TraversalInfo<NNFConvertData>& info)
/* ----------------------------------------------------------------------------
 *
 * Description:   Converts a propositional formula with either a Finally or a
 *                Globally operator to negation normal form, expanding the
 *                operator to an equivalent representation. Used by 
 *                FormulaString::toNegationNormalForm() when traversing a
 *                FormulaString to perform the conversion.
 *
 *                This function implements the conversions
 *                    Fp  ==>  true U p        Gp  ==>  false V p
 *                   !Fp  ==>  false V !p     !Gp  ==>  true U !p
 *                instantiated by the template parameters.
 *
 * Argument:      info  --  A reference to a TraversalInfo object containing
 *                          the information whether a negation needs to be
 *                          pushed over the operator, in addition with the
 *                          partially constructed result. The presence of a
 *                          negation affects the way the operator is expanded.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  if (info.operand == 0)
  {
    bool negate = info.data.negate_stack.top();
    *(info.data.result) += (negate ? OP2 : OP1);
    *(info.data.result) += ' ';
    *(info.data.result) += (negate == C) ? LTL_TRUE : LTL_FALSE;
    *(info.data.result) += ' ';
  }
}


 
/******************************************************************************
 *
 * Function definitions for class FormulaString::BinaryToNNF.
 *
 *****************************************************************************/

/* ========================================================================= */
template<char OP1, char OP2, bool B1, bool B2>
inline void FormulaString::BinaryToNNF::eval
  (FormulaString::TraversalInfo<NNFConvertData>& info)
/* ----------------------------------------------------------------------------
 *
 * Description:   Converts a propositional formula with a binary operator
 *                (other than an equivalence operator) to negation normal
 *                form. Used by 
 *                FormulaString::toNegationNormalForm() when traversing a
 *                FormulaString to perform the conversion.
 *
 *                This function implements the conversions
 *                  p1 /\ p2   ==>   p1 /\ p2        p1 \/ p2   ==>   p1 \/ p2
 *                !(p1 /\ p2)  ==>  !p1 \/ !p2     !(p1 \/ p2)  ==>  !p1 /\ !p2
 *
 *                  p1 -> p2   ==>  !p1 \/ p2      !(p1 -> p2)  ==>   p1 \/ !p2
 *
 *                   p1 U p2   ==>   p1 U p2          p1 V p2   ==>   p1 V p2
 *                 !(p1 U p2)  ==>  !p1 V !p2       !(p1 V p2)  ==>  !p1 U !p2
 *                instantiated by the template parameters.
 *
 * Argument:      info  --  A reference to a TraversalInfo object containing
 *                          the information whether a negation needs to be
 *                          pushed over the operator, in addition with the
 *                          partially constructed result. The presence of a
 *                          negation affects the way of rewriting the operator.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  if (info.operand == 0)
  {
    bool negate = info.data.negate_stack.top();
    
    info.data.negate_stack.top() = (negate ? B2 : !B2);
    info.data.negate_stack.push(negate ? B1 : !B1);
    
    *(info.data.result) += (negate ? OP2 : OP1);
    *(info.data.result) += ' ';
  }
}



/******************************************************************************
 *
 * Function definitions for class FormulaString::EquivalenceToNNF.
 *
 *****************************************************************************/

/* ========================================================================= */
void FormulaString::EquivalenceToNNF::eval
  (FormulaString::TraversalInfo<NNFConvertData>& info)
/* ----------------------------------------------------------------------------
 *
 * Description:   Converts a propositional formula with an equivalence operator
 *                to negation normal form. Used by 
 *                FormulaString::toNegationNormalForm() when traversing a
 *                FormulaString to perform the conversion.
 *
 *                This function implements the conversions
 *                    p1 <-> p2   ==>  (p1 \/ !p2) /\ (!p1 \/ p2)
 *                  !(p1 <-> p2)  ==>  (p1 /\ !p2) \/ (!p1 \/ p2) 
 *
 * Argument:      info  --  A reference to a TraversalInfo object containing
 *                          the information whether a negation needs to be
 *                          pushed over the operator, in addition with the
 *                          partially constructed result. The presence of a
 *                          negation affects the way of rewriting the operator.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  switch (info.operand)
  {
    case 0 :
      {
        bool negate = info.data.negate_stack.top();
        *(info.data.result) += (negate ? LTL_OR : LTL_AND);
        *(info.data.result) += ' ';
        *(info.data.result) += (negate ? LTL_AND : LTL_OR);
        *(info.data.result) += ' ';
        
        info.data.equivalence_operand_stack.push(info.data.result);
        info.data.result = new string;
        
        info.data.negate_stack.push(false);
      }
      break;
        
    case 1 :
      info.data.equivalence_operand_stack.push(info.data.result);
      info.data.result = new string;
      
      info.data.negate_stack.push(false);
      break;
        
    case 2 :
      {
        /*
         *  The handling of equivalences must be done in a special way.
         *  After both of the (unnegated) operands of the equivalence operator
         *  have been processed, the operand formulae are negated and converted
         *  to negation normal form independently by traversing both
         *  subformulae separately (the recursive calls to
         *  FormulaString::traverseStream). Then, the result is constructed
         *  from the four computed subformulae.
         */

        string* left_operand = info.data.equivalence_operand_stack.top();
        info.data.equivalence_operand_stack.pop();
        string* right_operand = info.data.result;

        /*
         *  Compute the negation of the left operand of the equivalence
         *  operator in negation normal form.
         */
        
        istrstream left_operand_stream(left_operand->c_str());
          
        info.data.result = new string;
        info.data.negate_stack.push(true);

        FormulaString::template
          traverseStream<AtomToNNF, FormulaString::ConstantToNNF<true>,
                         FormulaString::ConstantToNNF<false>, NegationToNNF,
                         NextToNNF,
                         FormulaString::AbbreviatedUnaryToNNF
                           <LTL_UNTIL, LTL_V, false>,
                         FormulaString::AbbreviatedUnaryToNNF
                           <LTL_V, LTL_UNTIL, true>,
                         FormulaString::BinaryToNNF<LTL_AND, LTL_OR, true,
                                                    true>,
                         FormulaString::BinaryToNNF<LTL_OR, LTL_AND, true,
                                                    true>,
                         FormulaString::BinaryToNNF<LTL_OR, LTL_AND, false,
                                                    true>,
                         EquivalenceToNNF,
                         FormulaString::BinaryToNNF<LTL_UNTIL, LTL_V, true, 
                                                    true>,
                         FormulaString::BinaryToNNF<LTL_V, LTL_UNTIL, true,
                                                    true>
                        >
          (PREORDER | INORDER | POSTORDER, left_operand_stream, info);

        string* neg_left_operand = info.data.result;
           
        /*
         *  Compute the negation of the right operand of the equivalence
         *  operator in negation normal form.
         */

        istrstream right_operand_stream(right_operand->c_str());
          
        info.data.result = new string;
        info.data.negate_stack.push(true);

        FormulaString::template
          traverseStream<AtomToNNF, FormulaString::ConstantToNNF<true>,
                         FormulaString::ConstantToNNF<false>, NegationToNNF,
                         NextToNNF,
                         FormulaString::AbbreviatedUnaryToNNF
                           <LTL_UNTIL, LTL_V, false>,
                         FormulaString::AbbreviatedUnaryToNNF
                           <LTL_V, LTL_UNTIL, true>,
                         FormulaString::BinaryToNNF<LTL_AND, LTL_OR, true,
                                                    true>,
                         FormulaString::BinaryToNNF<LTL_OR, LTL_AND, true,
                                                    true>,
                         FormulaString::BinaryToNNF<LTL_OR, LTL_AND, false,
                                                    true>,
                         EquivalenceToNNF,
                         FormulaString::BinaryToNNF<LTL_UNTIL, LTL_V, true,
                                                    true>,
                         FormulaString::BinaryToNNF<LTL_V, LTL_UNTIL, true,
                                                    true>
                        >
          (PREORDER | INORDER | POSTORDER, right_operand_stream, info);
        
        string* neg_right_operand = info.data.result;

        /*
         *  Construct the result.
         */

        info.operand = 2;

        info.data.result = info.data.equivalence_operand_stack.top();
        info.data.equivalence_operand_stack.pop();
      
        *(info.data.result) += *left_operand + *neg_right_operand
                               + (info.data.negate_stack.top() ? LTL_AND :
                                                                 LTL_OR)
                               + ' ' + *neg_left_operand + *right_operand;
            
        info.data.negate_stack.pop();
        if (info.data.negate_stack.empty())
          info.data.result->resize(info.data.result->size() - 1);
        
        delete left_operand;
        delete neg_left_operand;
        delete right_operand;
        delete neg_right_operand;
      } 
      break;
  }
}

}
