/*
 *  Copyright (C) 1999, 2000  Heikki Tauriainen <Heikki.Tauriainen@hut.fi>
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include "UserInterface.h"

/******************************************************************************
 *
 * Functions for displaying various statistics during testing.
 *
 *****************************************************************************/

/* ========================================================================= */
void UserInterface::changeStreamFormatting
  (ostream& stream, int width, int precision, ios::fmtflags flags)
/* ----------------------------------------------------------------------------
 *
 * Description:   Changes the formatting state of an output stream, storing its
 *                previous state so that it can be restored later.
 *
 * Arguments:     stream     --  A reference to an output stream.
 *                width      --  Field width.
 *                precision  --  Floating-point precision.
 *                flags      --  Flags affecting e.g. output justification.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  StreamFormatting formatting;

  formatting.width = stream.width(width);
  formatting.precision = stream.precision(precision);
  formatting.flags = stream.flags(flags);

  stream_formatting_stack.push(formatting);
}

/* ========================================================================= */
void UserInterface::restoreStreamFormatting(ostream& stream)
/* ----------------------------------------------------------------------------
 *
 * Description:   Restores the formatting state of an output stream whose
 *                previous state was saved by a call to
 *                UserInterface::changeStreamFormatting.
 *
 * Arguments:     stream  --  A reference to an output stream.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  if (!stream_formatting_stack.empty())
  {
    StreamFormatting formatting = stream_formatting_stack.top();
    stream_formatting_stack.pop();
    stream.width(formatting.width);
    stream.precision(formatting.precision);
    stream.flags(formatting.flags);
  }
}

/* ========================================================================= */
void UserInterface::printHeading
  (ostream& stream, int indent, const string& heading_text) const
/* ----------------------------------------------------------------------------
 *
 * Description:   Writes a line of text to a stream, e.g. a `heading' (used to
 *                explain what the following information is about when
 *                displaying messages).
 *
 * Arguments:     stream  --  A reference to an output stream.
 *                indent  --  Number of spaces to leave on the left of the
 *                            output.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  Exceptional_ostream estream(&stream, ios::failbit | ios::badbit);

  estream << string(indent, ' ') + heading_text << endl;
}

/* ========================================================================= */
void UserInterface::printBuchiAutomatonStats
  (ostream& stream, int indent,
   vector<AlgorithmTestResults, ALLOC(AlgorithmTestResults) >::size_type
     algorithm,
   int result_id)
/* ----------------------------------------------------------------------------
 *
 * Description:   Displays information about a Bchi automaton, extracting the
 *                information from a TestStatistics structure stored in the
 *                UserInterface object.
 *
 * Arguments:     stream      --  A reference to the output stream to which the
 *                                information is to be written.
 *                indent      --  Number of spaces to leave on the left of the
 *                                output.
 *                algorithm   --  Identifier of the algorithm used for
 *                                generating the automaton.
 *                result_id   --  Selects between the automata constructed
 *                                from a formula and its negation.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  Exceptional_ostream estream(&stream, ios::failbit | ios::badbit);

  const AutomatonStats& automaton_stats = 
    test_results[algorithm].automaton_stats[result_id];

  estream << string(indent, ' ');

  if (!automaton_stats.buchiAutomatonComputed())
    estream << "not computed";
  else
  {
    estream << "number of states:" + string(6, ' ')
            << automaton_stats.number_of_buchi_states
            << '\n' + string(indent, ' ') + "number of transitions: "
            << automaton_stats.number_of_buchi_transitions
            << '\n' + string(indent, ' ') + "computation time:"
               + string(6, ' ');
    
    if (automaton_stats.buchi_generation_time != -1.0)
    {
      changeStreamFormatting(stream, 9, 2, ios::fixed | ios::left);
      estream << automaton_stats.buchi_generation_time;
      restoreStreamFormatting(stream);

      estream << " seconds (user time)";
    }
    else
      estream << "unknown";
  }
  estream << endl;
}

/* ========================================================================= */
void UserInterface::printProductAutomatonStats
  (ostream& stream, int indent,
   vector<AlgorithmTestResults, ALLOC(AlgorithmTestResults) >::size_type
     algorithm,
   int result_id)
/* ----------------------------------------------------------------------------
 *
 * Description:   Displays information about a product automaton, extracting
 *                the information from a TestStatistics structure stored in the
 *                UserInterface object.
 *
 * Arguments:     stream     --  A reference to the output stream to which the
 *                               information is to be written.
 *                indent     --  Number of spaces to leave on the left of the
 *                               output.
 *                algorithm  --  Identifier of the algorithm used for
 *                               generating the product automaton.
 *                result_id  --  Selects between the automata constructed from
 *                               a formula and its negation.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  Exceptional_ostream estream(&stream, ios::failbit | ios::badbit);

  const AutomatonStats& automaton_stats =
    test_results[algorithm].automaton_stats[result_id];

  estream << string(indent, ' ');

  if (!automaton_stats.productAutomatonComputed())
    estream << "not computed";
  else
  {
    estream << "number of states:" + string(6, ' ');

    changeStreamFormatting(stream, 9, 0, ios::left);
    estream << automaton_stats.number_of_product_states;
    restoreStreamFormatting(stream);

    estream << " [";

    if (automaton_stats.number_of_product_states != 0)
    {
      changeStreamFormatting(stream, 0, 2, ios::fixed);
      estream << static_cast<double>
                   (automaton_stats.number_of_product_states)
                 / static_cast<double>(automaton_stats.number_of_buchi_states)
                 / static_cast<double>(round_info.statespace->size())
                 * 100.0;
      restoreStreamFormatting(stream);

      estream << "% of worst case ("
              << automaton_stats.number_of_buchi_states
                 * round_info.statespace->size()
              << ')';
    }
    else
      estream << "empty automaton";

    estream << "]\n" + string(indent, ' ') + "number of transitions: "
            << automaton_stats.number_of_product_transitions;
  }
  estream << endl;
}

/* ========================================================================= */
void UserInterface::printAcceptanceCycleStats
  (ostream& stream, int indent,
   vector<AlgorithmTestResults, ALLOC(AlgorithmTestResults) >::size_type
     algorithm,
   int result_id)
/* ----------------------------------------------------------------------------
 *
 * Description:   Displays information about the number of system states from
 *                which begins an execution accepted by a Bchi automaton. The
 *                information is extracted from a TestStatistics structure
 *                stored in the UserInterface object.
 *
 * Arguments:     stream      --  A reference to the output stream to which the
 *                                information is to be written.
 *                indent      --  Number of spaces to leave on the left of the
 *                                output.
 *                algorithm   --  Identifier of the algorithm used for
 *                                computing the Bchi automaton whose accepting
 *                                cycles are to be considered.
 *                result_id   --  Selects between the result computed for a
 *                                formula and its negation.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  Exceptional_ostream estream(&stream, ios::failbit | ios::badbit);

  const AutomatonStats& automaton_stats =
    test_results[algorithm].automaton_stats[result_id];

  estream << string(indent, ' ');

  if (!automaton_stats.emptinessCheckPerformed())
    estream << "not computed";
  else
  {
    estream << "cycle reachable from   ";

    changeStreamFormatting(stream, 9, 0, ios::left);
    estream << automaton_stats.emptiness_check_result->count();
    restoreStreamFormatting(stream);

    estream << " system states\n" + string(indent, ' ')
               + "not reachable from     ";

    changeStreamFormatting(stream, 9, 0, ios::left);
    estream << automaton_stats.emptiness_check_result->zeroCount();
    restoreStreamFormatting(stream);

    estream << " system states";
  }

  estream << endl;
}

/* ========================================================================= */
void UserInterface::printConsistencyCheckStats
  (ostream& stream, int indent,
   vector<AlgorithmTestResults, ALLOC(AlgorithmTestResults) >::size_type
     algorithm)
/* ----------------------------------------------------------------------------
 *
 * Description:   Displays information about the consistency check result for
 *                a given algorithm, extracting the information from a
 *                TestStatistics structure stored in the UserInterface object.
 *
 * Arguments:     stream     --  A reference to an output stream to which the
 *                               information is to be written.
 *                indent     --  Number of spaces to leave on the left of the
 *                               output.
 *                algorithm  --  Identifier of the algorithm whose consistency
 *                               check result is to be displayed.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  Exceptional_ostream estream(&stream, ios::failbit | ios::badbit);
  const AlgorithmTestResults& test_result = test_results[algorithm];

  estream << string(indent, ' ');

  if (test_result.consistency_check_result == -1)
    estream << "not performed";
  else
  {
    estream << "result:" + string(18, ' ');

    if (test_result.consistency_check_result == 0)
    {
      estream << "failed    ["
              << test_result.failed_consistency_check_comparisons
              << " (";

      changeStreamFormatting(stream, 0, 2, ios::fixed);
      estream << ((test_result.consistency_check_comparisons == 0)
                  ? 0.0
                  : static_cast<double>
                     (test_result.failed_consistency_check_comparisons)
                    / test_result.consistency_check_comparisons * 100.0);
      restoreStreamFormatting(stream);

      estream << "%) of " << test_result.consistency_check_comparisons
              << " test cases]";
    }
    else
      estream << "passed";
  }

  estream << endl;
}

/* ========================================================================= */
void UserInterface::printPathCheckStats
  (ostream& stream, int indent,
   vector<AlgorithmTestResults, ALLOC(AlgorithmTestResults) >::size_type
     algorithm) const
/* ----------------------------------------------------------------------------
 *
 * Description:   Displays information about the path check result for
 *                a given algorithm, extracting the information from a
 *                TestStatistics structure stored in the UserInterface object.
 *
 * Arguments:     stream     --  A reference to an output stream to which the
 *                               information is to be written.
 *                indent     --  Number of spaces to leave on the left of the
 *                               output.
 *                algorithm  --  Identifier of the algorithm whose consistency
 *                               check result is to be displayed.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  Exceptional_ostream estream(&stream, ios::failbit | ios::badbit);
  const AlgorithmTestResults& test_result = test_results[algorithm];

  for (int counter = 0; counter < 2; counter++)
  {
    estream << string(indent, ' ') + (counter == 1 ? "negated" : "positive")
               + " formula:" + string(8 + counter, ' ');

    if (test_result.automaton_stats[counter].path_check_result == -1)
      estream << "N/A";
    else
      estream << (test_result.automaton_stats[counter].path_check_result == 1
		  ? "pass" : "fail") << "ed";

    estream << endl;
  }
}

/* ========================================================================= */
void UserInterface::printCrossComparisonStats
  (ostream& stream, int indent,
   vector<AlgorithmTestResults, ALLOC(AlgorithmTestResults) >::size_type
     current_algorithm,
   vector<AlgorithmTestResults, ALLOC(AlgorithmTestResults) >::size_type
     last_algorithm)
/* ----------------------------------------------------------------------------
 *
 * Description:   Displays information about the cross-comparison check result
 *                for a given algorithm, extracting the information from
 *                a vector of TestStatistics structures stored in the
 *                UserInterface object.
 *
 * Arguments:     stream             --  A reference to an output stream to
 *                                       which the information is to be
 *                                       written.
 *                indent             --  Number of spaces to leave on the left
 *                                       of the output.
 *                current_algorithm  --  Identifier of the algorithm whose
 *                                       model checking results are to be
 *                                       compared with those of other
 *                                       algorithms.
 *                last_algorithm     --  Largest identifier of an algorithm
 *                                       whose model checking results are
 *                                       ready.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  if (last_algorithm == 0)
    return;

  Exceptional_ostream estream(&stream, ios::failbit | ios::badbit);
  const AlgorithmTestResults& test_result = test_results[current_algorithm];

  bool first_printed;

  for (int counter = 0; counter < 2; counter++)
  {
    first_printed = false;

    estream << string(indent, ' ') + (counter == 1
                                      ? "negated formula: "
                                      : "positive formula:")
               + string(8, ' ');

    for (vector<AlgorithmTestResults, ALLOC(AlgorithmTestResults) >::size_type
	   algorithm_id = 0;
         algorithm_id < last_algorithm; ++algorithm_id)
    {
      if (algorithm_id != current_algorithm)
      {
        if (first_printed)
          estream << string(indent + 25, ' ');
        else
          first_printed = true;

        if (!test_result.automaton_stats[counter].
              crossComparisonPerformed(algorithm_id))
          estream << "N/A:     ";
        else if (test_result.automaton_stats[counter].
                   cross_comparison_stats[algorithm_id].second == 0)
          estream << "ok:" + string(6, ' ');
        else
          estream << "mismatch:";

        estream << " `" + *(configuration.algorithms[algorithm_id].name)
                   + "'\n";

        if (!test_result.automaton_stats[counter].
              crossComparisonPerformed(algorithm_id)
            && test_result.automaton_stats[counter].
                 cross_comparison_stats[algorithm_id].second > 0)
        {
          estream << string(indent + 35, ' ') + '['
                  << test_result.automaton_stats[counter].
                       cross_comparison_stats[algorithm_id].second
                  << " (";

          changeStreamFormatting(stream, 0, 2, ios::fixed);
          estream << static_cast<double>
                       (test_result.automaton_stats[counter].
                          cross_comparison_stats[algorithm_id].second)
                     / test_result.automaton_stats[counter].
                         emptiness_check_result->size()
                     * 100.0;
          restoreStreamFormatting(stream);
        
          estream << "%) of " << test_result.automaton_stats[counter].
                                   emptiness_check_result->size()
                  << " test cases]\n";
        }
      }
    }
  }

  estream << endl;
}

/* ========================================================================= */
void UserInterface::printAllStats
  (ostream& stream, int indent,
   vector<AlgorithmTestResults, ALLOC(AlgorithmTestResults) >::size_type
     algorithm)
/* ----------------------------------------------------------------------------
 *
 * Description:   Displays all test information (Bchi automaton and product
 *                automaton statistics, acceptance cycle information,
 *                consistency check result, cross-comparison results) for an
 *                algorithm.
 *
 * Arguments:     stream     --  A reference to an output stream to which the
 *                               information is to be written.
 *                indent     --  Number of spaces to leave on the left of the
 *                               output.
 *                algorithm  --  Identifier of an algorithm.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  printHeading(stream, indent,
               '`' + *(configuration.algorithms[algorithm].name) + '\'');

  for (int counter = 0; counter < 2; counter++)
  {
    printHeading(stream, indent + 2,
                 string(counter == 0 ? "Positive" : "Negated")
                 + " formula:");
    
    printHeading(stream, indent + 4, "Bchi automaton:");
    printBuchiAutomatonStats(stream, indent + 6, algorithm, counter);
    printHeading(stream, indent + 4, "Product automaton:");
    printProductAutomatonStats(stream, indent + 6, algorithm, counter);
    printHeading(stream, indent + 4, "Acceptance cycles:");
    printAcceptanceCycleStats(stream, indent + 6, algorithm, counter);
  }

  printHeading(stream, indent + 2, "Consistency check:");
  printConsistencyCheckStats(stream, indent + 4, algorithm);

  if (configuration.statespace_options.generate_mode & Configuration::PATH)
  {
    printHeading(stream, indent + 2, "Path check result:");
    printPathCheckStats(stream, indent + 4, algorithm);
  }

  if (test_results.size() > 1)
  {
    printHeading(stream, indent + 2, "Result comparison:");
    printCrossComparisonStats(stream, indent + 4, algorithm,
                              test_results.size());
  }
  else
    stream << endl;
}

/* ========================================================================= */
void UserInterface::printCollectiveCrossComparisonStats
  (ostream& stream,
   vector<TestStatistics, ALLOC(TestStatistics) >::size_type algorithm_y,
   vector<TestStatistics, ALLOC(TestStatistics) >::size_type algorithm_x)
/* ----------------------------------------------------------------------------
 *
 * Description:   Called by UserInterface::printCollectiveStats in order to
 *                fill a single cell of the result cross-comparison table.
 *
 * Arguments:     stream                    --  A reference to an output
 *                                              stream.
 *                algorithm_x, algorithm_y  --  Identifiers of the algorithms
 *                                              whose cross-comparison results
 *                                              are to be displayed.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{        
  Exceptional_ostream estream(&stream, ios::failbit | ios::badbit);

  estream << ' ';
  if (algorithm_x == algorithm_y)
    estream << string(21, ' ');
  else if (final_statistics[algorithm_y].
             cross_comparisons_performed[algorithm_x] > 0)
  {
    changeStreamFormatting(stream, 5, 0, ios::right);
    estream << final_statistics[algorithm_y].
                 cross_comparison_mismatches[algorithm_x];
    restoreStreamFormatting(stream);

    estream << '/';

    changeStreamFormatting(stream, 5, 0, ios::left);
    estream << final_statistics[algorithm_y].
                 cross_comparisons_performed[algorithm_x];
    restoreStreamFormatting(stream);

    estream << " (";

    double percentage = static_cast<double>
                          (final_statistics[algorithm_y].
                             cross_comparison_mismatches[algorithm_x])
                        / final_statistics[algorithm_y].
                          cross_comparisons_performed[algorithm_x]
                        * 100.0;

    changeStreamFormatting(stream, 0, 2, ios::fixed);
    estream << percentage;
    restoreStreamFormatting(stream);
      
    estream << "%)";

    if (percentage < 100.0)
      estream << ' ';    
    if (percentage < 10.0)
      estream << ' ';
  }
  else
    estream << "[none performed]" + string(5, ' ');
}

/* ========================================================================= */
void UserInterface::printCollectiveStats
  (ostream& stream, int indent,
   unsigned long int round,
   unsigned long int generated_statespaces,
   unsigned long int generated_formulae)
/* ----------------------------------------------------------------------------
 *
 * Description:   Displays average information about a series of tests for each
 *                tested algorithm:
 *                  - average sizes of the Bchi and product automata
 *                  - number of failed attempts to generate a Bchi automaton
 *                  - number of failed consistency checks
 *                  - cross-comparison mismatches
 *
 * Arguments:     stream                 --  A reference to an output stream.
 *                indent                 --  Number of spaces to leave on the
 *                                           left of the output.
 *                round                  --  Number of performed test rounds.
 *                generated_statespaces  --  Number of generated state spaces.
 *                generated_formulae     --  Number of generated LTL formulae.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  Exceptional_ostream estream(&stream, ios::failbit | ios::badbit);
  const string ind(indent, ' ');

  estream << '\n' + ind + "Statistics after round " << round << ":\n\n";

  estream << ind + "  Number of generated state spaces: "
          << generated_statespaces
          << '\n' + ind + "  Number of generated LTL formulae: "
          << generated_formulae
          << '\n';

  string algorithm_name;

  for (int i = 0;
       i < (configuration.statespace_options.generate_mode
	      & Configuration::GRAPH
	    ? 3 : 4);
       i++)
  {
    estream << '\n' + string(2 + indent, ' ');

    switch (i)
    {
      case 1 :
        estream << "Failures to compute Bchi automaton\n"
                   + string(2 + indent, ' ') + string(35, '=') + '\n';
        break;
      case 2 :
        estream << "Consistency check failures\n" + string(2 + indent, ' ')
                   + string(26, '=') + '\n';
        break;
      case 3 :
	estream << "Path check failures\n" + string(2 + indent, ' ')
    	           + string(19, '=') + '\n';
	break;
      default :
        break;
    }

    for (vector<Configuration::AlgorithmInformation,
	        ALLOC(Configuration::AlgorithmInformation) >::size_type
           algorithm = 0;
         algorithm != configuration.algorithms.size();
         ++algorithm)
    {
      estream << '\n' + string((i > 0 ? 4 : 2) + indent, ' ')
                 + *(configuration.algorithms[algorithm].name) + '\n';

      switch (i)
      {
        case 0 :
        {
          unsigned long int failures_to_compute_automaton;
          unsigned long int automaton_count;
          unsigned long int number_of_successful_instances;
          unsigned long int total_number_of_states;
          unsigned long int total_number_of_transitions;
          double buchi_generation_time;

          const TestStatistics& stats = final_statistics[algorithm];

          estream << string(2 + indent,  ' ')
                     + string(configuration.algorithms[algorithm].name
                            ->length(), '=');

          for (int k = 0; k < 2; k++)
          {
            estream << "\n\n" + string(8 + indent, ' ')
                       + (k == 0 ? "BCHI  " : "PRODUCT") + string(7, ' ')
                       + "|  Number of  |   States    | Transitions |"
                       + (k == 0 ? "  Avg. time  |\n" : "\n")
                       + string(8 + indent, ' ')
                       + "AUTOMATA" + string(6, ' ')
                       + "|  automata   | / automaton | / automaton |"
                         + (k == 0 ? "  (seconds)  |\n" : "\n")
                       + string(7 + indent, ' ') + "--";

            for (int j = k; j < 5; j++)
              estream << string(13, '-') + '+';

            estream << '\n';

            for (int j = 0; j < 3; j++)
            {
              estream << string(8 + indent, ' ');
              switch (j)
              {
                case 0  : estream << "Pos. formulae"; break;
                case 1  : estream << "Neg. formulae"; break;
                default : estream << "All formulae "; break;
              }

              estream << " | ";

              if (j < 2)
              {
                failures_to_compute_automaton =
                  stats.failures_to_compute_automaton[j];

                automaton_count =
                  stats.automaton_count[j];

                total_number_of_states =
                  (k == 0 ? stats.total_number_of_buchi_states[j]
                          : stats.total_number_of_product_states[j]);

                total_number_of_transitions =
                  (k == 0 ? stats.total_number_of_buchi_transitions[j]
                          : stats.total_number_of_product_transitions[j]);

                buchi_generation_time =
                  (k == 0 ? stats.total_buchi_generation_time[j] : -1.0);
              }
              else
              {
                failures_to_compute_automaton =
                  stats.failures_to_compute_automaton[0]
                  + stats.failures_to_compute_automaton[1];

                automaton_count =
                  stats.automaton_count[0] + stats.automaton_count[1];

                total_number_of_states =
                  (k == 0 ? stats.total_number_of_buchi_states[0]
                            + stats.total_number_of_buchi_states[1]
                          : stats.total_number_of_product_states[0]
                            + stats.total_number_of_product_states[1]);

                total_number_of_transitions =
                  (k == 0 ? stats.total_number_of_buchi_transitions[0]
                            + stats.total_number_of_buchi_transitions[1]
                          : stats.total_number_of_product_transitions[0]
                            + stats.total_number_of_product_transitions[1]);

                buchi_generation_time =
                  (k == 0 && stats.total_buchi_generation_time[0] >= 0.0
                   && stats.total_buchi_generation_time[1] >= 0.0
                   ? stats.total_buchi_generation_time[0]
                     + stats.total_buchi_generation_time[1]
                   : -1.0);
              }

              number_of_successful_instances =
                automaton_count - failures_to_compute_automaton;

              if (number_of_successful_instances == 0)
                estream << (k == 0 ? "gen. failed" : "uncomputed ");
              else 
              {
                changeStreamFormatting(stream, 11, 2, ios::right);
                estream << number_of_successful_instances;
                restoreStreamFormatting(stream);
              }

              estream << " | ";

              if (number_of_successful_instances == 0)
                estream << string(11, ' ');
              else
              {
                changeStreamFormatting(stream, 11, 2, ios::fixed | ios::right);
                estream << total_number_of_states
                           / static_cast<double>
                               (number_of_successful_instances);
                restoreStreamFormatting(stream);
              }

              estream << " | ";

              if (number_of_successful_instances == 0)
                estream << string(11, ' ');
              else
              {
                changeStreamFormatting(stream, 11, 2, ios::fixed | ios::right);
                estream << total_number_of_transitions
                           / static_cast<double>
                               (number_of_successful_instances);
                restoreStreamFormatting(stream);
              }

              if (buchi_generation_time >= 0.0)
              {
                estream << " | ";

                if (number_of_successful_instances == 0)
                  estream << string(11, ' ');
                else
                {
                  changeStreamFormatting(stream, 11, 2,
                                         ios::fixed | ios::right);
                  estream << buchi_generation_time
                             / number_of_successful_instances;
                  restoreStreamFormatting(stream);
                }
              }

              estream << " |\n";
            }
          }
        
          break;
        }

        case 1 :
        {
          unsigned long int failures_to_compute_automaton;
          unsigned long int automaton_count;

          const TestStatistics& stats = final_statistics[algorithm];

          for (int j = 0; j < 3; j++)
          {
            estream << string(8 + indent, ' ');
            switch (j)
            {
              case 0  : estream << "Positive formulae: "; break;
              case 1  : estream << "Negative formulae: "; break;
              default : estream << "Total:" + string(13, ' '); break;
            }

            if (j < 2)
            {
              failures_to_compute_automaton =
                stats.failures_to_compute_automaton[j];
              automaton_count =
                stats.automaton_count[j];
            }
            else
            {
              failures_to_compute_automaton =
                stats.failures_to_compute_automaton[0]
                + stats.failures_to_compute_automaton[1];

              automaton_count = 
                stats.automaton_count[0] + stats.automaton_count[0];
            }

            changeStreamFormatting(stream, 5, 0, ios::left);
            estream << failures_to_compute_automaton;
            restoreStreamFormatting(stream);

            estream << " [";

	    if (automaton_count > 0)
	    {
	      changeStreamFormatting(stream, 0, 2, ios::fixed);
	      estream << static_cast<double>(failures_to_compute_automaton)
                         / automaton_count * 100.0;
	      restoreStreamFormatting(stream);

	      estream << "% of " << automaton_count;
	    }
	    else
	      estream << "no";

	    estream << " attempts]\n";
          }

          break;
        }

        case 2 :
        {
          const TestStatistics& stats = final_statistics[algorithm];
          
          estream << string(8 + indent, ' ');

          changeStreamFormatting(stream, 5, 0, ios::left);
          estream << stats.consistency_check_failures;
          restoreStreamFormatting(stream);

          estream << " [";

          if (stats.consistency_checks_performed > 0)
          {
            changeStreamFormatting(stream, 0, 2, ios::fixed);
            estream << static_cast<double>(stats.consistency_check_failures)
                       / stats.consistency_checks_performed 
                       * 100.0;
            restoreStreamFormatting(stream);

            estream << "% of " << stats.consistency_checks_performed;
          } 
          else
            estream << "no";

          estream << " checks performed]\n";

          break;
        }

        case 3 :
	{
	  unsigned long int path_check_failures;
          unsigned long int path_check_count;

          const TestStatistics& stats = final_statistics[algorithm];

          for (int j = 0; j < 3; j++)
          {
            estream << string(8 + indent, ' ');
            switch (j)
            {
              case 0  : estream << "Positive formulae: "; break;
              case 1  : estream << "Negative formulae: "; break;
              default : estream << "Total:" + string(13, ' '); break;
            }

            if (j < 2)
            {
              path_check_failures = stats.path_check_failures[j];
              path_check_count = stats.path_checks_performed[j];
            }
            else
            {
              path_check_failures = stats.path_check_failures[0]
                                    + stats.path_check_failures[1];

              path_check_count = stats.path_checks_performed[0]
                                 + stats.path_checks_performed[1];
            }

            changeStreamFormatting(stream, 5, 0, ios::left);
            estream << path_check_failures;
            restoreStreamFormatting(stream);

            estream << " [";

	    if (path_check_count > 0)
	    {
	      changeStreamFormatting(stream, 0, 2, ios::fixed);
	      estream << static_cast<double>(path_check_failures)
                         / path_check_count * 100.0;
	      restoreStreamFormatting(stream);

	      estream << "% of " << path_check_count;
	    }
	    else
	      estream << "no";

	    estream << " checks performed]\n";
          }

          break;
	}
      }
    }
    estream << '\n';
  }
  
  if (final_statistics.size() > 1)
  {
    estream << '\n' + ind + "  Cross-comparison mismatches "
                            "[failures / comparisons performed]\n"
               + ind + "  " + string(62, '=') + '\n';

    vector<TestStatistics, ALLOC(TestStatistics) >::size_type
      algorithm_x, algorithm_y;

    for (algorithm_x = 0; algorithm_x < final_statistics.size();
         algorithm_x += 2)
    {
      estream << '\n' + string(25 + indent, ' ');

      for (int i = 0; i < 2; ++i)
      {
        if (algorithm_x + i < final_statistics.size())
        {
          algorithm_name =
            (*(configuration.algorithms[algorithm_x + i].name)).substr(0, 20);
          estream << "| " + algorithm_name
                     + string(21 - algorithm_name.length(), ' ');
        }
      }
      estream << '|';

      for (algorithm_y = 0; algorithm_y < final_statistics.size();
           algorithm_y++)
      {
        estream << "\n    " + ind + string(21, '-');

        for (int i = 0; i < 2; ++i)
        {
          if (algorithm_x + i < final_statistics.size())
            estream << '+' + string(22, '-');
        }
        estream << '+';

        algorithm_name
          = (*(configuration.algorithms[algorithm_y].name)).substr(0, 20);

        estream << "\n    " + ind + algorithm_name
                   + string(21 - algorithm_name.length(), ' ');
        
        for (int i = 0; i < 2; ++i)
        {
          if (algorithm_x + i  < final_statistics.size())
          {
            estream << '|';
            printCollectiveCrossComparisonStats(stream, algorithm_y,
                                                algorithm_x + i);
          }
        }
        estream << '|';
      }

      estream << '\n';
    }
  }

  estream << endl;
  estream.flush();
}
