/*
 *  Copyright (C) 1999, 2000  Heikki Tauriainen <Heikki.Tauriainen@hut.fi>
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#ifndef _TestOperations_h_
#define _TestOperations_h_

#include <cstdio>
#include <ctime>
#include <sys/times.h>
#include <fstream>
#include <iostream>
#include <strstream>
#include <vector>
#include "Bitset.h"
#include "BuchiAutomaton.h"
#include "Configuration.h"
#include "Exception.h"
#include "FormulaString.h"
#include "ProductAutomaton.h"
#include "StateSpace.h"
#include "StlConfig.h"
#include "TestRoundInfo.h"
#include "TestStatistics.h"

using namespace std;

extern Configuration configuration;

using Graph::BuchiAutomaton;
using Graph::StateSpace;
using Graph::ProductAutomaton;



/******************************************************************************
 *
 * Prototypes for functions implementing various test operations.
 *
 *****************************************************************************/

void openFile
  (const char* filename, ifstream& stream,          /* Opens a file for */
   ios::openmode mode, int indent = 0);             /* reading.         */

void openFile                                       /* Opens a file for */
  (const char* filename, ofstream& stream,          /* writing.         */
   ios::openmode mode, int indent = 0);

void removeFile                                     /* Removes a file. */
  (const char* filename, int indent = 0);

void printFileContents                              /* Displays the contents */
  (ostream& stream, const char* message,            /* of a file.            */
   const char* filename, int indent,
   const char* line_prefix = "");

StateSpace* generateStateSpace();                   /* Generates a 
						     * reachability graph.
						     */

void generateFormulae                               /* Generates a random   */
  (FormulaString::FormulaString formulae[4],        /* LTL formula and      */
   const char* formula = 0);                        /* stores the formula,
						     * its negation and the
						     * NNFs of the two
						     * formulae into a
						     * given array.
						     */

void writeFormulaToFile                             /* Writes a LTL formula */
  (const FormulaString::FormulaString& formula,     /* into a file.         */
   const char* filename);

BuchiAutomaton* generateBuchiAutomaton              /* Generates a Bchi     */
  (const char* formula_file_name,                   /* automaton from a LTL  */
   const Configuration::AlgorithmInformation&       /* formula stored in a   */
     algorithm,                                     /* given file, using a   */
   AutomatonStats& automaton_stats, int indent);    /* given LTL-to-Bchi
						     * translation algorithm
						     * for the conversion.
						     */

ProductAutomaton* generateProductAutomaton          /* Computes the         */
  (const BuchiAutomaton& automaton,                 /* synchronous product  */
   const StateSpace& statespace,                    /* of a Bchi automaton */
   AutomatonStats& automaton_stats, int indent);    /* and a reachability
						     * graph.
						     */

void performEmptinessCheck                          /* Performs an emptiness */
  (const ProductAutomaton& product_automaton,       /* check on a product    */
   AutomatonStats& automaton_stats,                 /* automaton.            */
   StateSpace::size_type result_size);

bool performConsistencyCheck                        /* Performs a           */
  (AlgorithmTestResults& test_results);             /* consistency check on
						     * the test results
						     * for a formula and its
						     * negation.
						     */

bool performTruthValueCheck                         /* Compares the model   */
  (const AlgorithmTestResults& test_results,        /* checking results     */
   const Bitset& positive_formula_truth_values,     /* obtained using some  */
   const Bitset& negative_formula_truth_values);    /* LTL-to-Bchi
						     * conversion algorithm
						     * with the results
						     * obtained by
						     * evaluating the
						     * formula directly.
						     */

bool compareResults                                 /* Compares the model   */
  (vector<AlgorithmTestResults,                     /* checking results     */
          ALLOC(AlgorithmTestResults) >&            /* obtained using some  */
     test_results,                                  /* LTL-to-Bchi         */
   vector<AlgorithmTestResults,                     /* conversion algorithm */
          ALLOC(AlgorithmTestResults) >::size_type  /* with the results     */
     current_algorithm,                             /* given by the other   */
   vector<TestStatistics, ALLOC(TestStatistics) >&  /* algorithms.          */
     stats);

bool performPathCheck                               /* Compares the model   */
  (const Bitset* path_truth_values,                 /* checking results     */
   AutomatonStats& test_result,                     /* obtained using some  */
   bool negate);				    /* LTL-to-Bchi
						     * conversion algorithm
						     * with the results
						     * obtained by
						     * evaluating the
						     * formula directly in
						     * a path.
						     */



/******************************************************************************
 *
 * A class for reporting state space generation errors.
 *
 *****************************************************************************/

class StateSpaceGenerationException : public Exception
{
public:
  StateSpaceGenerationException();                  /* Constructor. */

  /* default copy constructor */

  ~StateSpaceGenerationException();                 /* Destructor. */

  StateSpaceGenerationException& operator=          /* Assignment operator. */
    (const StateSpaceGenerationException& e);
};

/* ========================================================================= */
inline StateSpaceGenerationException::StateSpaceGenerationException() :
  Exception("State space generation failed.")
/* ----------------------------------------------------------------------------
 *
 * Description:   Constructor for class StateSpaceGenerationException.
 *
 * Arguments:     None.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
}

/* ========================================================================= */
inline StateSpaceGenerationException::~StateSpaceGenerationException()
/* ----------------------------------------------------------------------------
 *
 * Description:   Destructor for class StateSpaceGenerationException.
 *
 * Arguments:     None.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
}

/* ========================================================================= */
inline StateSpaceGenerationException&
StateSpaceGenerationException::operator=
  (const StateSpaceGenerationException& e)
/* ----------------------------------------------------------------------------
 *
 * Description:   Assignment operator for class StateSpaceGenerationException.
 *                Assigns the value of another StateSpaceGenerationException to
 *                `this' one.
 *
 * Arguments:     e  --  A reference to a constant StateSpaceException.
 *
 * Returns:       A reference to the object whose values was changed.
 *
 * ------------------------------------------------------------------------- */
{
  Exception::operator=(e);
  return *this;
}



/******************************************************************************
 *
 * A class for reporting Bchi automaton generation errors.
 *
 *****************************************************************************/

class BuchiAutomatonGenerationException : public Exception
{
public:
  BuchiAutomatonGenerationException();              /* Constructor. */

  /* default copy constructor */

  ~BuchiAutomatonGenerationException();             /* Destructor. */

  BuchiAutomatonGenerationException& operator=      /* Assignment operator. */
    (const BuchiAutomatonGenerationException& e);
};

/* ========================================================================= */
inline BuchiAutomatonGenerationException::BuchiAutomatonGenerationException() :
  Exception("Bchi automaton generation failed.")
/* ----------------------------------------------------------------------------
 *
 * Description:   Constructor for class BuchiAutomatonGenerationException.
 *
 * Arguments:     None.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
}

/* ========================================================================= */
inline BuchiAutomatonGenerationException::~BuchiAutomatonGenerationException()
/* ----------------------------------------------------------------------------
 *
 * Description:   Destructor for class BuchiAutomatonGenerationException.
 *
 * Arguments:     None.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
}

/* ========================================================================= */
inline BuchiAutomatonGenerationException&
BuchiAutomatonGenerationException::operator=
  (const BuchiAutomatonGenerationException& e)
/* ----------------------------------------------------------------------------
 *
 * Description:   Assignment operator for class
 *                BuchiAutomatonGenerationException. Assigns the value of
 *                another BuchiAutomatonGenerationException to `this' one.
 *
 * Arguments:     e  --  A reference to a constant
 *                       BuchiAutomatonGenerationException.
 *
 * Returns:       A reference to the object whose value was changed.
 *
 * ------------------------------------------------------------------------- */
{
  Exception::operator=(e);
  return *this;
}



/******************************************************************************
 *
 * A class for reporting product automaton generation errors.
 *
 *****************************************************************************/

class ProductAutomatonGenerationException : public Exception
{
public:
  ProductAutomatonGenerationException();            /* Constructor. */

  /* default copy constructor */

  ~ProductAutomatonGenerationException();           /* Destructor. */

  ProductAutomatonGenerationException& operator=    /* Assignment operator. */
    (const ProductAutomatonGenerationException& e);
};

/* ========================================================================= */
inline
ProductAutomatonGenerationException::ProductAutomatonGenerationException() :
  Exception("Product automaton generation failed.")
/* ----------------------------------------------------------------------------
 *
 * Description:   Constructor for class ProductAutomatonGenerationException.
 *
 * Arguments:     None.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
}

/* ========================================================================= */
inline
ProductAutomatonGenerationException::~ProductAutomatonGenerationException()
/* ----------------------------------------------------------------------------
 *
 * Description:   Destructor for class ProductAutomatonGenerationException.
 *
 * Arguments:     None.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
}

/* ========================================================================= */
inline ProductAutomatonGenerationException&
ProductAutomatonGenerationException::operator=
  (const ProductAutomatonGenerationException& e)
/* ----------------------------------------------------------------------------
 *
 * Description:   Assignment operator for class
 *                ProductAutomatonGenerationException.
 *
 * Arguments:     e  --  A reference to a constant
 *                       ProductAutomatonGenerationException.
 *
 * Returns:       A reference to the object whose value was changed.
 *
 * ------------------------------------------------------------------------- */
{
  Exception::operator=(e);
  return *this;
}



/******************************************************************************
 *
 * A class for reporting errors during the emptiness check.
 *
 *****************************************************************************/

class EmptinessCheckFailedException : public Exception
{
public:
  EmptinessCheckFailedException();                  /* Constructor. */

  /* default copy constructor */

  ~EmptinessCheckFailedException();                 /* Destructor. */

  EmptinessCheckFailedException& operator=          /* Assignment operator. */
    (const EmptinessCheckFailedException& e);
};

/* ========================================================================= */
inline
EmptinessCheckFailedException::EmptinessCheckFailedException() :
  Exception("Emptiness check failed.")
/* ----------------------------------------------------------------------------
 *
 * Description:   Constructor for class EmptinessCheckFailedException.
 *
 * Arguments:     None.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
}

/* ========================================================================= */
inline EmptinessCheckFailedException::~EmptinessCheckFailedException()
/* ----------------------------------------------------------------------------
 *
 * Description:   Destructor for class EmptinessCheckFailedException.
 *
 * Arguments:     None.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
}

/* ========================================================================= */
inline EmptinessCheckFailedException&
EmptinessCheckFailedException::operator=
  (const EmptinessCheckFailedException& e)
/* ----------------------------------------------------------------------------
 *
 * Description:   Assignment operator for class EmptinessCheckFailedException.
 *                Assigns the value of another EmptinessCheckFailedException to
 *                `this' one.
 *
 * Arguments:     e  --  A reference to a constant
 *                       EmptinessCheckFailedException.
 *
 * Returns:       A reference to the object whose value was changed.
 *
 * ------------------------------------------------------------------------- */
{
  Exception::operator=(e);
  return *this;
}



#endif /* _TestOperations_h_ */
