/*
 *  Copyright (C) 1999, 2000  Heikki Tauriainen <Heikki.Tauriainen@hut.fi>
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#ifndef _TestStatistics_h_
#define _TestStatistics_h_

#include <utility>
#include <vector>
#include "BuchiAutomaton.h"
#include "Configuration.h"
#include "ProductAutomaton.h"
#include "StateSpace.h"
#include "StlConfig.h"

using namespace std;
using Graph::BuchiAutomaton;
using Graph::StateSpace;
using Graph::ProductAutomaton;

/******************************************************************************
 *
 * A data structure for storing test data for a single formula.
 *
 *****************************************************************************/

struct AutomatonStats
{
  explicit AutomatonStats                           /* Constructor. */
    (vector<Configuration::AlgorithmInformation,
            ALLOC(Configuration::AlgorithmInformation) >
     ::size_type number_of_algorithms);

  /* default copy constructor */

  ~AutomatonStats();                                /* Destructor. */

  /* default assignment operator */

  void reset();                                     /* Resets the statistics.
                                                     */

  bool buchiAutomatonComputed() const;              /* Tests whether the
                                                     * statistics indicate that
                                                     * a Bchi automaton has
                                                     * been computed.
						     */

  bool productAutomatonComputed() const;            /* Tests whether the
                                                     * statistics indicate that
                                                     * a product automaton has
                                                     * been computed.
						     */

  bool emptinessCheckPerformed() const;             /* Tests whether the
                                                     * statistics indicate that
                                                     * the emptiness check has
                                                     * been performed for a
                                                     * product automaton.
						     */

  bool crossComparisonPerformed                     /* Tests whether the     */
    (unsigned long int algorithm) const;            /* statistics indicate that
                                                     * the result of the
                                                     * emptiness check has been
                                                     * compared against the
                                                     * result of a given
                                                     * algorithm.
						     */

  BuchiAutomaton::size_type number_of_buchi_states; /* Number of states in a
                                                     * Bchi automaton.
						     */

  unsigned long int number_of_buchi_transitions;    /* Number of transitions in
                                                     * a Bchi automaton.
						     */

  unsigned long int number_of_msccs;                /* Number of maximal
                                                     * strongly connected
                                                     * components in the
                                                     * automaton.
                                                     */

  double buchi_generation_time;                     /* Time used to generate a
                                                     * Bchi automaton.
						     */

  ProductAutomaton::size_type                       /* Number of stats in a */
    number_of_product_states;                       /* product automaton.   */

  unsigned long int number_of_product_transitions;  /* Number of transitions in
                                                     * a product automaton.
						     */

  Bitset* emptiness_check_result;                   /* Pointer to the result of
                                                     * the emptiness check for
                                                     * the product automaton.
                                                     */

  int path_check_result;                            /* Tells the result of the
						     * path check for an
						     * algorithm. The value
						     * -1 means that no check
						     * was performed, while
						     * value 0 denotes a failed
						     * check and the value 1
						     * represents a successful
						     * check. (The path check
						     * result is used only when
						     * reachability graphs are
						     * single paths.)
						     */

  typedef pair<bool, unsigned long int>
    StatPair;

  vector<StatPair, ALLOC(StatPair) >                /* Emptiness check       */
    cross_comparison_stats;                         /* cross-comparison
                                                     * results. The `first'
                                                     * element of the pair
                                                     * tells whether a cross-
                                                     * comparison with a given
                                                     * algorithm has been
                                                     * performed, and the
                                                     * `second' element of the
                                                     * pair gives the number
                                                     * of system states in
                                                     * which the results
                                                     * differ.
						     */
};



/******************************************************************************
 *
 * A data structure for storing test data for a single algorithm.
 *
 *****************************************************************************/

struct AlgorithmTestResults
{
  explicit AlgorithmTestResults                     /* Constructor. */
    (vector<Configuration::AlgorithmInformation,
            ALLOC(Configuration::AlgorithmInformation) >
       ::size_type
       number_of_algorithms);

  /* default copy constructor */

  ~AlgorithmTestResults();                          /* Destructor. */

  /* default assignment operator */

  void reset();                                     /* Resets the test results.
                                                     */

  int consistency_check_result;                     /* Tells the consistency
                                                     * check status for an
                                                     * algorithm. The value
                                                     * -1 means the check has
                                                     * not been performed, a 0
                                                     * stands for a failed
                                                     * check, and the value 1
                                                     * denotes that the check
                                                     * was successful.
						     */

  StateSpace::size_type                             /* Number of test cases */
    consistency_check_comparisons;                  /* in the consistency
                                                     * check.
						     */

  StateSpace::size_type                             /* Number of failed test */
    failed_consistency_check_comparisons;           /* cases in the consistency
                                                     * check.
						     */

  vector<AutomatonStats, ALLOC(AutomatonStats) >    /* A two-element vector */
    automaton_stats;                                /* storing test results
                                                     * for an algorithm.
						     */
};



/******************************************************************************
 *
 * A data structure for storing test statistics for a single algorithm over the
 * whole test session.
 *
 *****************************************************************************/

struct TestStatistics
{
  explicit TestStatistics                           /* Constructor. */
    (vector<TestStatistics,
            ALLOC(TestStatistics) >::size_type
       number_of_algorithms);

  /* default copy constructor */

  ~TestStatistics();

  /* default assignment operator */

  unsigned long int                                 /* Number of failed      */
    failures_to_compute_automaton[2];               /* attempts to generate
                                                     * a Bchi automaton.
						     */

  unsigned long int automaton_count[2];             /* Number of attempts to
                                                     * generate a Bchi
                                                     * automaton.
						     */

  unsigned long int consistency_check_failures;     /* Number of failed
                                                     * consistency checks.
						     */

  unsigned long int consistency_checks_performed;   /* Number of consistency
                                                     * checks performed.
						     */

  unsigned long int path_check_failures[2];         /* Number of failed path
                              			     * checks.
						     */

  unsigned long int path_checks_performed[2];       /* Number of path checks
                                                     * performed.
						     */

  unsigned long int                                 /* Total number of   */
    total_number_of_buchi_states[2];                /* states in all the
                                                     * generated Bchi
                                                     * automata.
						     */

  unsigned long int                                 /* Total number of      */
    total_number_of_buchi_transitions[2];           /* transitions in all the
                                                     * generated Bchi
                                                     * automata.
						     */

  unsigned long int                                 /* Total number of      */
    total_number_of_msccs[2];                       /* maximal strongly
						     * connected components
                                                     * in the generated
                                                     * Bchi automata.
						     */

  double total_buchi_generation_time[2];            /* Total time used when
                                                     * generating Bchi
                                                     * automata.
						     */

  unsigned long int                                 /* Total number of   */
    total_number_of_product_states[2];              /* states in all the 
                                                     * generated product
                                                     * automata.
						     */

  unsigned long int                                 /* Total number of      */
    total_number_of_product_transitions[2];         /* transitions in all the
                                                     * generated product
                                                     * automata.
						     */

  vector<unsigned long int,                         /* Number of failed */
         ALLOC(unsigned long int) >                 /* result cross-    */
    cross_comparison_mismatches;                    /* comparisons.     */

  vector<unsigned long int,                         /* Number of result  */
         ALLOC(unsigned long int) >                 /* cross-comparisons */
    cross_comparisons_performed;                    /* performed.        */
};



/******************************************************************************
 *
 * Inline function definitions for struct AutomatonStats.
 *
 *****************************************************************************/

/* ========================================================================= */
inline AutomatonStats::AutomatonStats
  (vector<Configuration::AlgorithmInformation,
          ALLOC(Configuration::AlgorithmInformation) >::size_type
     number_of_algorithms) :
  number_of_buchi_states(0), number_of_buchi_transitions(1),
  number_of_msccs(0), buchi_generation_time(0.0), number_of_product_states(0),
  number_of_product_transitions(1), emptiness_check_result(0),
  path_check_result(-1),
  cross_comparison_stats(number_of_algorithms, make_pair(false, 0))
/* ----------------------------------------------------------------------------
 *
 * Description:   Constructor for struct AutomatonStats.
 *
 * Arguments:     number_of_algorithms  --  Number of implementations taking
 *                                          part in the tests.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
}

/* ========================================================================= */
inline AutomatonStats::~AutomatonStats()
/* ----------------------------------------------------------------------------
 *
 * Description:   Destructor for struct AutomatonStats.
 *
 * Arguments:     None.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
}

/* ========================================================================= */
inline bool AutomatonStats::buchiAutomatonComputed() const
/* ----------------------------------------------------------------------------
 *
 * Description:   Test whether the statistics indicate that a Bchi automaton
 *                has been computed (i.e. whether information about the
 *                automaton has been stored in the AutomatonStats structure).
 *
 * Arguments:     None.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  return (number_of_buchi_states != 0
          || number_of_buchi_transitions <= number_of_buchi_states);
}

/* ========================================================================= */
inline bool AutomatonStats::productAutomatonComputed() const
/* ----------------------------------------------------------------------------
 *
 * Description:   Test whether the statistics indicate that a product automaton
 *                has been computed (i.e. whether information about the
 *                automaton has been stored in the AutomatonStats structure).
 *
 * Arguments:     None.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  return (number_of_product_states != 0
          || number_of_product_transitions <= number_of_product_states);
}

/* ========================================================================= */
inline bool AutomatonStats::emptinessCheckPerformed() const
/* ----------------------------------------------------------------------------
 *
 * Description:   Test whether the statistics indicate that an emptiness check
 *                has been performed for a product automaton (i.e. whether
 *                a valid pointer to the emptiness check result has been stored
 *                in the AutomatonStats structure).
 *
 * Arguments:     None.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  return (emptiness_check_result != 0);
}

/* ========================================================================= */
inline bool
AutomatonStats::crossComparisonPerformed(unsigned long int algorithm) const
/* ----------------------------------------------------------------------------
 *
 * Description:   Test whether the statistics indicate that the emptiness check
 *                result for a product automaton has been compared with another
 *                result produced using a different algorithm.
 *
 * Arguments:     None.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  return cross_comparison_stats[algorithm].first;
}



/******************************************************************************
 *
 * Inline function definitions for struct AlgorithmTestResults.
 *
 *****************************************************************************/

/* ========================================================================= */
inline AlgorithmTestResults::AlgorithmTestResults
  (vector<Configuration::AlgorithmInformation,
          ALLOC(Configuration::AlgorithmInformation) >::size_type
     number_of_algorithms) :
  consistency_check_result(-1), consistency_check_comparisons(0),
  failed_consistency_check_comparisons(0),
  automaton_stats(2, AutomatonStats(number_of_algorithms))
/* ----------------------------------------------------------------------------
 *
 * Description:   Constructor for struct AlgorithmTestResults.
 *
 * Arguments:     number_of_algorithms  --  Number of implementations taking
 *                                          part in the tests.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
}

/* ========================================================================= */
inline AlgorithmTestResults::~AlgorithmTestResults()
/* ----------------------------------------------------------------------------
 *
 * Description:   Destructor for struct AlgorithmTestResults.
 *
 * Arguments:     None.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
}



/******************************************************************************
 *
 * Inline function definitions for struct TestStatistics.
 *
 *****************************************************************************/

/* ========================================================================= */
inline TestStatistics::TestStatistics
  (vector<TestStatistics, ALLOC(TestStatistics) >::size_type
     number_of_algorithms) :
  consistency_check_failures(0), consistency_checks_performed(0),
  cross_comparison_mismatches(number_of_algorithms, 0),
  cross_comparisons_performed(number_of_algorithms, 0)
/* ----------------------------------------------------------------------------
 *
 * Description:   Constructor for struct TestStatistics.
 *
 * Arguments:     number_of_algorithms  --  Number of implementations taking
 *                                          part in the tests.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  for (int i = 0; i < 2; i++)
  {
    failures_to_compute_automaton[i] = 0;
    path_check_failures[i] = 0;
    path_checks_performed[i] = 0;
    automaton_count[i] = 0;
    total_number_of_buchi_states[i] = 0;
    total_number_of_buchi_transitions[i] = 0;
    total_number_of_msccs[i] = 0;
    total_number_of_product_states[i] = 0;
    total_number_of_product_transitions[i] = 0;
    total_buchi_generation_time[i] = 0.0;
  }
}

/* ========================================================================= */
inline TestStatistics::~TestStatistics()
/* ----------------------------------------------------------------------------
 *
 * Description:   Destructor for struct TestStatistics.
 *
 * Arguments:     None.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
}

#endif /* _TestStatistics_h_ */
