/*
 *  Copyright (C) 1999, 2000  Heikki Tauriainen <Heikki.Tauriainen@hut.fi>
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include "UserInterface.h"

/******************************************************************************
 *
 * Functions for reading and parsing user commands.
 *
 *****************************************************************************/

/* ========================================================================= */
void UserInterface::executeUserCommands
  (unsigned long int generated_statespaces,
   unsigned long int generated_formulae)
/* ----------------------------------------------------------------------------
 *
 * Description:   Loop for reading user commands and executing them after a
 *                test round.
 *
 * Arguments:     generated_statespaces  --  Number of generated state spaces.
 *                generated_formulae     --  Number of generated formulae.
 *
 * Returns:       Nothing. However, changes `round_info.current_round',
 *                `round_info.next_round_to_stop',
 *                `round_info.next_round_to_run' and `round_info.abort' as a
 *                side effect, depending on the user's wish to abort testing,
 *                skip some number of test rounds or continue testing for a
 *                number of rounds.
 *
 * ------------------------------------------------------------------------ */
{
  string input_line;
  vector<string, ALLOC(string) > input_tokens;
  TokenType token;

  bool formula_type = true;

  pair<string, bool> redirection_info;
  ofstream output_file;
  ostream* output_stream = 0;

  int indent;

  BuchiAutomaton* buchi_automaton = 0;
  pair<unsigned long int, bool> last_computed_buchi_automaton;

  ProductAutomaton* product_automaton = 0;
  pair<unsigned long int, bool> last_computed_product_automaton;

  while (1)
  {
    cout << "  ** [Round "
	 << round_info.current_round
         << '/'
         << configuration.global_options.number_of_rounds
         << "] >> ";

    cout.flush();

    try
    {
      getline(cin, input_line, '\n');
      
      sliceString(input_line, " \t", input_tokens);
      
      if (!input_tokens.empty())
      {
        cout << endl;

        token = parseCommand(input_tokens[0]);

        if (token == CONTINUE || token == SKIP)
        {
          verifyArgumentCount(input_tokens, 0, 1);

          unsigned long int rounds_to_continue;

          rounds_to_continue = (input_tokens.size() > 1
                                ? parseNumber(input_tokens[1])
                                : 1);
          
          if (rounds_to_continue == 0)
            throw CommandErrorException("Argument of the command must be "
                                        "positive.");

          if (input_tokens.size() == 1
              && token == CONTINUE
              && configuration.global_options.interactive
                   != Configuration::ALWAYS)
	  {
	    bool all_algorithms_disabled = true;

	    for (vector<Configuration::AlgorithmInformation,
                        ALLOC(Configuration::AlgorithmInformation) >
		   ::const_iterator
		   algorithm = configuration.algorithms.begin();
		 algorithm != configuration.algorithms.end();
		 ++algorithm)
	    {
	      if (algorithm->enabled)
	      {
		all_algorithms_disabled = false;
		break;
	      }
	    }

	    if (all_algorithms_disabled)
	    {
	      cout << "  Warning! All algorithms are currently disabled.\n"
		      "  Are you sure you wish to continue? [y/n] ";
	      cout.flush();

	      getline(cin, input_line, '\n');
	      sliceString(input_line, " \t", input_tokens);

	      if (!input_tokens.empty() && (input_tokens[0][0] == 'y'
					    || input_tokens[0][0] == 'Y'))
		round_info.next_round_to_stop
		  = configuration.global_options.number_of_rounds + 1;
	      else
	      {
		cout << endl;
		continue;
	      }
	    }
	  }
          else
            round_info.next_round_to_stop
              = round_info.current_round + rounds_to_continue;

          if (token == SKIP)
            round_info.next_round_to_run = round_info.next_round_to_stop + 1;

          break;
        }
	else if (token == DISABLE)
	{
	  verifyArgumentCount(input_tokens, 0, 1);
	  changeAlgorithmState(input_tokens, false);
	  continue;
	}
	else if (token == ENABLE)
	{
	  verifyArgumentCount(input_tokens, 0, 1);
	  changeAlgorithmState(input_tokens, true);
	  continue;
	}
        else if (token == QUIT)
        {
          verifyArgumentCount(input_tokens, 0, 0);
          round_info.abort = 1;
          break;
        }
        else if (token == VERBOSITY)
        {
          verifyArgumentCount(input_tokens, 0, 1);
          changeVerbosity(input_tokens);
          continue;
        }

	if (round_info.skip && (token == ANALYZE || token == BUCHI ||
				token == EVALUATE || token == INCONSISTENCIES))
	  throw CommandErrorException("Command cannot be performed since "
				      "this round was skipped.");

        /*
         *  If the command expects a formula identifier as a parameter, 
         *  determine the type of the formula which the command refers to.
         */

        if (token == ANALYZE || token == BUCHI || token == EVALUATE ||
            token == FORMULA || token == PATHINCONSISTENCIES)
          formula_type = parseFormulaType(input_tokens);

        /*
         *  Determine whether the output of the command is to be saved or
         *  appended to a file, instead of displaying it on the console. If
         *  output redirection is required, open a file for output.
         */

        redirection_info = parseRedirection(input_tokens);

        if (redirection_info.first.empty())
        {
          output_stream = &cout;
          indent = 2;
        }
        else
        {
          openFile(redirection_info.first.c_str(), output_file,
                   ios::out | (redirection_info.second
                               ? ios::app
                               : ios::trunc),
                   2);

          output_stream = &output_file;
          indent = 0;
        }

        switch (token)
        {
          case ALGORITHMS :
            verifyArgumentCount(input_tokens, 0, 0);
            printAlgorithmList(*output_stream, indent);
            if (output_stream != &cout)
              cout << "  List of algorithms";
            break;

          case ANALYZE :
            verifyArgumentCount(input_tokens, 2, 2);
            printAnalysisResults(*output_stream, indent,
                                 formula_type,
                                 parseNumber(input_tokens[1]),
                                 parseNumber(input_tokens[2]),
                                 buchi_automaton,
                                 last_computed_buchi_automaton,
                                 product_automaton,
                                 last_computed_product_automaton);
            if (output_stream != &cout)
              cout << "  Cross-comparison mismatch analysis result";
            break;

          case BUCHI :
            verifyArgumentCount(input_tokens, 1, 2);
            printBuchiAutomaton(*output_stream, indent,
                                formula_type,
                                input_tokens,
                                buchi_automaton,
                                last_computed_buchi_automaton,
                                (input_tokens.size() == 3
                                 && input_tokens[2] == "dot"
                                 ? Graph::DOT
                                 : Graph::NORMAL));
            if (output_stream != &cout)
            {
              cout << "  Bchi automaton information";
              if (input_tokens.size() == 3 && input_tokens[2] == "dot")
                cout << " (in dot format)";
            }
            break;

          case EVALUATE :
            verifyArgumentCount(input_tokens, 0, 2);
            evaluateFormula(*output_stream, indent, formula_type,
                            input_tokens);
            if (output_stream != &cout)
              cout << "  Formula acceptance information";
            break;

          case FORMULA :
            verifyArgumentCount(input_tokens, 0, 0);
            printFormula(*output_stream, indent, formula_type);
            if (output_stream != &cout)
              cout << "  " << (formula_type ? "Formula" : "Negated formula");
            break;

          case HELP :
            verifyArgumentCount(input_tokens, 0, 1);
            printCommandHelp(*output_stream, indent, input_tokens);
            if (output_stream != &cout)
              cout << "  Command help";
            break;

          case INCONSISTENCIES :
            verifyArgumentCount(input_tokens, 0, 1);
            printInconsistencies(*output_stream, indent,
                                 round_info.current_round, input_tokens);
            if (output_stream != &cout)
              cout << "  Consistency check results for round "
                   << round_info.current_round;
            break;

	  case PATHINCONSISTENCIES :
	    if (configuration.statespace_options.generate_mode
		  & Configuration::GRAPH)
	      throw CommandErrorException("Command available only when using "
					  "paths as reachability graphs.");

	    verifyArgumentCount(input_tokens, 0, 1);
	    printPathInconsistencies(*output_stream, indent, formula_type,
				     round_info, input_tokens);
	    if (output_stream != &cout)
	      cout << "  Path check results for round "
		   << round_info.current_round;
	    break;

          case RESULTS :
            verifyArgumentCount(input_tokens, 0, 1);
            printTestResults(*output_stream, indent, round_info.current_round, 
                             input_tokens);
            if (output_stream != &cout)
              cout << "  Test results for round " << round_info.current_round;
            break;

          case STATESPACE :
            verifyArgumentCount(input_tokens, 0, 1);
            printStateSpace(*output_stream, indent, input_tokens,
                            (input_tokens.size() == 2 
                             && input_tokens[1] == "dot"
                             ? Graph::DOT
                             : Graph::NORMAL));
            if (output_stream != &cout)
            {
              cout << "  Reachability graph information";
              if (input_tokens.size() == 2 && input_tokens[1] == "dot")
                cout << " (in dot format)";
            }
            break;

          case STATISTICS :
            verifyArgumentCount(input_tokens, 0, 0);
            printCollectiveStats(*output_stream, indent,
                                 round_info.current_round,
                                 generated_statespaces, generated_formulae);
            if (output_stream != &cout)
              cout << "  Test statistics after round "
                   << round_info.current_round;
            break;

          default :
            throw CommandErrorException("Unknown command (`"
                                        + input_tokens[0] + "').");
        }

        if (output_stream != &cout)
        {
          output_file.close();
          cout << string(redirection_info.second ? " appended" : " written")
                  + " to `" + redirection_info.first + "'.\n\n";
        }
      }
    }
    catch (Exception e)
    {
      cout << string("  Error: ") + e.what() + "\n\n";
      if (output_stream != &cout && output_file.is_open())
        output_file.close();
    }
  }

  if (buchi_automaton != 0)
  {
    if (configuration.global_options.verbosity >= 4)
      cout << "  <cleaning up memory allocated for Bchi automaton>" << endl;

    delete buchi_automaton;
  }
  
  if (product_automaton != 0)
  {
    if (configuration.global_options.verbosity >= 4)
      cout << "  <cleaning up memory allocated for product automaton>"
           << endl;

    delete product_automaton;
  }
}      

/* ========================================================================= */
void UserInterface::printTextBlock
  (ostream& stream, int indent, const string& text, int max_line_len)
/* ----------------------------------------------------------------------------
 *
 * Description:   Writes a string of text to a stream, breaking the text to
 *                multiple indented lines of a given maximum length if
 *                required.
 *
 * Arguments:     stream        --  A reference to an output stream.
 *                indent        --  Number of spaces to leave on the left of
 *                                  each printed line.
 *                text          --  Text to be written to the stream. '\n':s
 *                                  can be used in the text to insert
 *                                  additional line breaks to the output. In
 *                                  this case, the '\n':s must be separated
 *                                  from the rest of the text by spaces.
 *                max_line_len  --  Maximum allowed line length (including the
 *                                  amount of indentation).
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  Exceptional_ostream estream(&stream, ios::failbit | ios::badbit);
  string::size_type max_len, current_pos = 1;
  vector<string, ALLOC(string) > words;
  const string ind = string(indent, ' ');

  if (indent > max_line_len)
    max_len = 0;
  else
    max_len = max_line_len - indent;

  sliceString(text, " ", words);

  if (words.empty())
    return;

  estream << ind;

  for (vector<string, ALLOC(string) >::const_iterator word = words.begin();
       word != words.end();
       ++word)
  {
    if (word->length() > max_len)
    {
      estream << *word + '\n';
      current_pos = 1;

      if (word != words.end() - 1)
        estream << ind;
    }
    else
    {
      current_pos += word->length();

      if (current_pos > max_len)
      {
        estream << '\n' + ind;
        current_pos = word->length() + 1;
      }
      
      estream << *word;

      if (word != words.end() - 1)
      {
        if (*word == "\n")
        {
          current_pos = 1;
          estream << ind;
        }
        else
        {
          estream << ' ';
          ++current_pos;
        }
      }
      else
        estream << '\n';
    }
  }
}

/* ========================================================================= */
void UserInterface::sliceString
  (const string& s, const char* slice_chars,
   vector<string, ALLOC(string) >& slices)
/* ----------------------------------------------------------------------------
 *
 * Description:   Slices a string into a vector of strings, using a given set
 *                of characters as separators.
 *
 * Arguments:     s            --  A reference to the constant original string.
 *                slice_chars  --  A C-style string containing the characters
 *                                 to be used as separators.
 *                slices       --  A reference to a vector for storing the
 *                                 string components.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  string::size_type last_non_slicechar_pos = 0;
  string::size_type last_slicechar_pos = 0;

  slices.clear();

  do
  {
    last_non_slicechar_pos = 
      s.find_first_not_of(slice_chars, last_slicechar_pos);
    if (last_non_slicechar_pos != s.npos)
    {
      last_slicechar_pos = s.find_first_of(slice_chars,
                                           last_non_slicechar_pos);
      if (last_slicechar_pos == s.npos)
        slices.push_back(s.substr(last_non_slicechar_pos));
      else
        slices.push_back(s.substr(last_non_slicechar_pos,
                                  last_slicechar_pos 
                                  - last_non_slicechar_pos));
    }
  }
  while (last_non_slicechar_pos != s.npos && last_slicechar_pos != s.npos);
}

/* ========================================================================= */
UserInterface::TokenType UserInterface::parseCommand(const string& token)
/* ----------------------------------------------------------------------------
 *
 * Description:   Parses a user command by translating a command name into
 *                its corresponding TokenType identifier.
 *
 * Argument:      token  --  A reference to a string containing the command.
 *
 * Returns:       A command identifier of the enumerated type TokenType.
 *
 * ------------------------------------------------------------------------- */
{
  TokenType token_type = UNKNOWN;
  string::size_type len = token.length();

  if (token.empty())
    return token_type;

  switch (token[0])
  {
    case 'a' :
      if (len < 2)
        throw CommandErrorException("Ambiguous command.");
        
      if (token[1] == 'l')
      {
        if (token.compare("gorithms", 2, len - 2) == 0)
          token_type = ALGORITHMS;
      }
      else if (token[1] == 'n')
      {
        if (token.compare("alyze", 2, len - 2) == 0)
          token_type = ANALYZE;
      }

      break;

    case 'b' :
      if (token.compare("uchi", 1, len - 1) == 0)
        token_type = BUCHI;
      break;

    case 'c' :
      if (token.compare("ontinue", 1, len - 1) == 0)
        token_type = CONTINUE;
      break;

    case 'd' :
      if (token.compare("isable", 1, len - 1) == 0)
	token_type = DISABLE;
      break;

    case 'e' :
      if (len < 2)
	throw CommandErrorException("Ambigous command.");

      if (token[1] == 'n')
      {
	if (token.compare("able", 2, len - 2) == 0)
	  token_type = ENABLE;
      }
      else if (token[1] == 'v')
      {
	if (token.compare("aluate", 2, len - 2) == 0)
	  token_type = EVALUATE;
      }

      break;

    case 'f' :
      if (token.compare("ormula", 1, len - 1) == 0)
        token_type = FORMULA;
      break;

    case 'h' :
      if (token.compare("elp", 1, len - 1) == 0)
        token_type = HELP;
      break;

    case 'i' :
      if (token.compare("nconsistencies", 1, len - 1) == 0)
        token_type = INCONSISTENCIES;
      break;

    case 'p' :
      if (token.compare("athinconsistencies", 1, len - 1) == 0)
	token_type = PATHINCONSISTENCIES;
      break;

    case 'q' :
      if (token.compare("uit", 1, len - 1) == 0)
        token_type = QUIT;
      break;

    case 'r' :
      if (token.compare("esults", 1, len - 1) == 0)
        token_type = RESULTS;
      break;

    case 's' :
      if (len < 2)
        throw CommandErrorException("Ambiguous command.");
      if (token[1] == 'k')
      {
        if (token.compare("ip", 2, len - 2) == 0)
          token_type = SKIP;
      }
      else if (token[1] == 't')
      {
        if (len < 5)
          throw CommandErrorException("Ambiguous command.");
        if (token[2] == 'a' && token[3] == 't')
        {
          if (token[4] == 'e')
          {
            if (token.compare("space", 5, len - 5) == 0)
              token_type = STATESPACE;
          }
          else if (token[4] == 'i')
          {
            if (token.compare("stics", 5, len - 5) == 0)
              token_type = STATISTICS;
          }
        }
      }
      break;

    case 'v' :
      if (token.compare("erbosity", 1, len - 1) == 0)
        token_type = VERBOSITY;
      break;
  }

  return token_type;
}

/* ========================================================================= */
void UserInterface::verifyArgumentCount
  (const vector<string, ALLOC(string) >& command,
   vector<string, ALLOC(string) >::size_type min_arg_count,
   vector<string, ALLOC(string) >::size_type max_arg_count)
/* ----------------------------------------------------------------------------
 *
 * Description:   Verifies that the number of arguments given for a user
 *                command is between a given interval.
 *
 * Arguments:     command        --  A reference to a constant vector of
 *                                   strings (the user command and its
 *                                   arguments).
 *                min_arg_count  --  Smallest allowed number of arguments.
 *                max_arg_count  --  Largest allowed number of arguments.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  if (command.size() < min_arg_count + 1)
    throw CommandErrorException("Too few arguments for command.");
  else if (command.size() > max_arg_count + 1)
    throw CommandErrorException("Too many arguments for command.");
}

/* ========================================================================= */
pair<string, bool> UserInterface::parseRedirection
  (vector<string, ALLOC(string) >& input_tokens)
/* ----------------------------------------------------------------------------
 *
 * Description:   Tests whether the last argument to a user command specifies
 *                output redirection. If redirection is desired, the 
 *                `argument' specifying the redirection is removed from the
 *                vector of strings forming the command.
 *
 * Argument:      input_tokens  --  A reference to a vector of strings giving
 *                                  the user command and its arguments.
 *
 * Returns:       A pair whose first component is the name of the output file
 *                (or the empty string if no redirection was specified) and
 *                whose second component determines whether the output is to
 *                be appended to the file instead of creating a new file.
 *
 * ------------------------------------------------------------------------- */
{
  string filename;
  bool append = false;

  if (!input_tokens.empty())
  {
    string& token = input_tokens.back();

    if (token[0] == '>')
    {
      if (token.length() > 1)
      {
        if (token[1] == '>')
        {
          if (token.length() > 2)
          {
            append = true;
            filename = token.substr(2);
          }
        }
        else
          filename = token.substr(1);
      }
      input_tokens.pop_back();
    }
    else if (input_tokens.size() >= 2)
    {
      string& token = *(input_tokens.rbegin() + 1);

      if (token[0] == '>' && (token.length() == 1
                              || (token.length() == 2 && token[1] == '>')))
      {
        filename = input_tokens.back();
        append = (token.length() == 2);
        input_tokens.pop_back();
        input_tokens.pop_back();
      }
    }
  }

  return make_pair(filename, append);
}

/* ========================================================================= */
bool UserInterface::parseFormulaType
  (vector<string, ALLOC(string) >& input_tokens)
/* ----------------------------------------------------------------------------
 *
 * Description:   Tests whether the first argument of a command specifies a
 *                formula (i.e. whether the first argument of the command is
 *                either a `+' or a `-'). If it is, the argument is removed
 *                from the vector of strings forming the command.
 *
 * Argument:      input_tokens  --  A reference to a vector of strings giving
 *                                  the user command.
 *
 * Returns:       A truth value according to whether a formula or its negation
 *                was specified; the effect of specifying no formula type is
 *                the same as giving a `+' as an argument (i.e. the formula
 *                type defaults to the positive formula).
 *
 * ------------------------------------------------------------------------- */
{
  bool formula_type = true;

  if (input_tokens.size() >= 2)
  {
    formula_type = (input_tokens[1] != "-");

    if (input_tokens[1].length() == 1
        && (input_tokens[1][0] == '+' || input_tokens[1][0] == '-'))
      input_tokens.erase(input_tokens.begin());
  }

  return formula_type;
}

/* ========================================================================= */
unsigned long int UserInterface::parseNumber(const string& number_string)
/* ----------------------------------------------------------------------------
 *
 * Description:   Converts a string to an unsigned long integer.
 *
 * Argument:      number_string  --  A reference to a constant string.
 *
 * Returns:       The number contained in the string, unless the string could
 *                not be converted to a number, in which case an exception is
 *                thrown.
 *
 * ------------------------------------------------------------------------- */
{
  char* endptr;
  unsigned long int number = strtoul(number_string.c_str(), &endptr, 10);
                
  if (*endptr != '\0')
    throw CommandErrorException("Expected a nonnegative integer, got `"
                                + number_string + "'.");

  return number;
}

/* ========================================================================= */
void UserInterface::parseInterval
  (const string& token,
   set<unsigned long int, less<unsigned long int>, ALLOC(unsigned long int) >& 
     number_set,
   unsigned long int min, unsigned long int max)
/* ----------------------------------------------------------------------------
 *
 * Description:   Parses a string for a list of number intervals, storing all
 *                the numbers into a set.
 *
 * Arguments:     token       --  A reference to a constant string containing
 *                                the list of intervals. A legal list of 
 *                                intervals has the following syntax:
 *
 *                                  <interval_list>
 *                                     ::=   <number>
 *                                         | '*'          // all numbers in the
 *                                                        //   interval
 *                                                        //   [min, max]
 *                                         | '-'<number>  // all numbers in the
 *                                                        //   interval
 *                                                        //   [min, number]
 *                                         | <number>'-'  // all numbers in the
 *                                                        //   interval
 *                                                        //   [number, max]
 *                                         | <number>'-'<number>
 *                                         | <interval_list>','<interval_list>
 *                            
 *                number_set  --  A reference to a set of unsigned long
 *                                integers for storing the result.
 *                min         --  Minimum bound for the numbers.
 *                max         --  Maximum bound for the numbers.
 *
 * Returns:       Nothing.
 *
 * ------------------------------------------------------------------------- */
{
  vector<string, ALLOC(string) > intervals;

  sliceString(token, ",", intervals);
  string::size_type dash_pos;

  number_set.clear();

  for (vector<string, ALLOC(string) >::const_iterator
	 interval = intervals.begin();
       interval != intervals.end();
       ++interval)
  {
    if (*interval == "*")
    {
      for (unsigned long int i = min; i <= max; i++)
        number_set.insert(i);
      break;
    }

    dash_pos = (*interval).find_first_of("-");

    if (dash_pos == (*interval).npos)
      number_set.insert(parseNumber(*interval));
    else
    {
      if (dash_pos == 0)
      {
        unsigned long int number = parseNumber((*interval).substr(1));
        for (unsigned long int i = min; i <= number; i++)
          number_set.insert(i);
      }
      else if (dash_pos == (*interval).length() - 1)
      {
        unsigned long int number =
          parseNumber((*interval).substr(0, (*interval).length() - 1));
        for (unsigned long int i = number; i <= max; i++)
          number_set.insert(i);
      }
      else
      {
        unsigned long int min_bound =
          parseNumber((*interval).substr(0, dash_pos));
        unsigned long int max_bound =
          parseNumber((*interval).substr(dash_pos + 1));

        for (unsigned long int i = min_bound; i <= max_bound; i++)
          number_set.insert(i);
      }
    }
  }
}
