// This file is based on code originally written by Mauno Rnkk,
// Copyright (C) 1998 Mauno Rnkk <mronkko@abo.fi>.
// Minor modifications by Heikki Tauriainen <Heikki.Tauriainen@hut.fi>.
//
// The original version of this file can be obtained through
//   <URL: http://www.abo.fi/~mauno.ronkko/ASA/ltlalg.html>.
//  
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.


#include "Ltl.h"


/// Used by function newName().
static int g_newName=LtlGraph::Node::INITNAME+1;



/// ----------Methods for LtlFormula----------

LtlFormula::LtlFormula()
{ 
  m_ltl=new Ltl(); 
}
  

LtlFormula::LtlFormula (const Ltl &formula)
{ 
  m_ltl=formula.clone(); 
}

  
LtlFormula::LtlFormula (const LtlFormula &ltl)
{ 
  m_ltl=ltl.m_ltl->clone(); 
}
  

LtlFormula::~LtlFormula()
{ 
  delete m_ltl; 
}
  

const bool LtlFormula::operator == (const LtlFormula &ltl) const
{ 
  return *m_ltl == *ltl.m_ltl; 
}
  

const bool LtlFormula::operator != (const LtlFormula &ltl) const
{ 
  return *m_ltl != *ltl.m_ltl; 
}
  

const LtlFormula& LtlFormula::operator = (const LtlFormula &ltl)
{
  Ltl *temp=ltl.m_ltl->clone();

  delete m_ltl;
  m_ltl=temp;

  return *this;
}


ostream& operator << (ostream& str, const LtlFormula& ltl)
{ 
  return str << *ltl.m_ltl; 
}




/// ----------Methods for LtlGraph::Node----------

LtlGraph::Node::Node()
{
  m_name      = INITNAME;
  m_father    = INITNAME;
}


LtlGraph::Node::Node (const LtlGraph::Node &node)
{
  m_name      = node.m_name;
  m_father    = node.m_father;
  m_incoming  = node.m_incoming;
  m_new       = node.m_new;
  m_old       = node.m_old;
  m_atomic    = node.m_atomic;
  m_next      = node.m_next;
}


LtlGraph::Node::Node (const int name, 
		      const int father, 
		      const Set<int> &incoming,
		      const Set<LtlFormula> &neww,
		      const Set<LtlFormula> &old,
		      const Set<LtlFormula> &atomic,
		      const Set<LtlFormula> &next)
{
  m_name      = name;
  m_father    = father;
  m_incoming  = incoming;
  m_new       = neww;
  m_old       = old;
  m_atomic    = atomic;
  m_next      = next;
}


const LtlGraph::Node& LtlGraph::Node::operator = (const LtlGraph::Node &node)
{
  m_name      = node.m_name;
  m_father    = node.m_father;
  m_incoming  = node.m_incoming;
  m_new       = node.m_new;
  m_old       = node.m_old;
  m_atomic    = node.m_atomic;
  m_next      = node.m_next;
  return *this;
}


const bool LtlGraph::Node::operator == (const LtlGraph::Node& node) const
{
  return  ((m_name==node.m_name) &&
	   (m_father==node.m_father) &&
	   (m_incoming==node.m_incoming) &&
	   (m_new==node.m_new) &&
	   (m_old==node.m_old) &&
	   (m_atomic==node.m_atomic) &&
	   (m_next==node.m_next));
}


const bool sameOldAndNext (const LtlGraph::Node &a, const LtlGraph::Node &c)
{ 
  return (a.m_old==c.m_old) && (a.m_next==c.m_next); 
}


const bool sameName (const LtlGraph::Node &a, const LtlGraph::Node &c)
{ 
  return (a.m_name==c.m_name);
}


ostream& operator << (ostream& str, const LtlGraph::Node& node)
{ 
  str << endl;
  str << " [ name    =" << node.m_name << endl; 
  str << "   father  =" << node.m_father << endl; 
  str << "   incoming=" << node.m_incoming << endl; 
  str << "   new     =" << node.m_new << endl; 
  str << "   old     =" << node.m_old << endl; 
  str << "   atomic  =" << node.m_atomic << endl; 
  str << "   next    =" << node.m_next << endl; 
  return str << " ] "; 
}




/// ----------Methods for LtlGraph----------

int newName()
{
  return g_newName++;
}


LtlGraph::LtlGraph (const Ltl &formula)
{
  int newname=newName();

  *this=formula.expand(Node(newname,
			    newname,
			    Set<int>(Node::INITNAME),
			    Set<LtlFormula>(LtlFormula(formula)),
			    Set<LtlFormula>(),
			    Set<LtlFormula>(),
			    Set<LtlFormula>()),
		       LtlGraph());
}


LtlGraph::LtlGraph (const LtlFormula &formula)
{
  int newname=newName();

  *this=formula.ltl()->expand(Node(newname,
				   newname,
				   Set<int>(Node::INITNAME),
				   Set<LtlFormula>(LtlFormula(formula)),
				   Set<LtlFormula>(),
				   Set<LtlFormula>(),
				   Set<LtlFormula>()),
			      LtlGraph());
}


ostream& operator << (ostream& str, const LtlGraph& set)
{ 
  Set<int> states;
  int      state;

  states=set.states();
  while (!states.isEmpty()) {
    state=*states.some();
    if (state==set.initialState())
      str << "initial";
    else
      str << state << ": " << set.propositions(state);
    str << " -> " << set.nextStates(state) << endl;
    states=states-state;
  }
  return str; 
}


const Set<int> LtlGraph::states() const
{
  Set<int>  result;
  Set<Node> temp(m_nodes);

  if (!m_nodes.isEmpty()){
    result=Node::INITNAME;
    while (!temp.isEmpty()) {
      result = result+temp.some()->m_name;
      temp   = temp-*temp.some();
    }
  }
  return result;
}


const Set<LtlFormula> LtlGraph::propositions(int state) const
{
  Node            criteria,*node;
  Set<LtlFormula> result;

  criteria.m_name=state;
  node=m_nodes.matching(sameName, criteria);
  if (node!=NULL)
    result=node->m_atomic;
  return result;
}


const Set<LtlFormula> LtlGraph::temporalProperties(int state) const
{
  Node            criteria,*node;
  Set<LtlFormula> result;

  criteria.m_name=state;
  node=m_nodes.matching(sameName, criteria);
  if (node!=NULL)
    result=node->m_old;
  return result;
}
  

const Set<int> LtlGraph::nextStates (int state) const
{
  Set<int>  result;
  Set<Node> temp(m_nodes);
  Node      *node;

  while (!temp.isEmpty()) {
    node=temp.some();
    if (node->m_incoming >= state)
      result=result + node->m_name;
    temp=temp-*node;
  }
  return result;
}





/// ----------Methods for Ltl----------

const LtlGraph Ltl::expand (const LtlGraph::Node &node, 
			    const LtlGraph &set) const
{
  LtlFormula     eta;
  LtlGraph::Node *nd;
  LtlGraph       newset;
  int            newname;

  if (node.m_new.isEmpty()) {
    newset=set;
    nd=newset.whichSameOldAndNext (node);
    if (nd!=NULL) {
      nd->m_incoming = nd->m_incoming + node.m_incoming;
      return newset;
    } else {
      newname=newName();
      return expand(LtlGraph::Node(newname, 
				   newname,
				   Set<int>(node.m_name),
				   node.m_next,
				   Set<LtlFormula>(),
				   Set<LtlFormula>(),
				   Set<LtlFormula>()),
		    LtlGraph(set.m_nodes + node));
    }
  } else {
    eta=*node.m_new.some();
    return eta.ltl()->expand1 (LtlGraph::Node(node.m_name, 
					      node.m_father,
					      node.m_incoming,
					      node.m_new - eta,
					      node.m_old + eta,
					      node.m_atomic,
					      node.m_next),
			       set);
  }
}


ostream& operator << (ostream& str, const Ltl& formula)
{ 
  return str << formula.expr(); 
}




/// ----------Methods for LtlNot----------

LtlNot::LtlNot (const Ltl &formula)
{ 
  m_formula=formula.clone(); 
  m_propagated=formula.negClone();
}


LtlNot::~LtlNot()
{ 
  delete m_formula; 
  delete m_propagated;
}




/// ----------Methods for LtlFalse----------

Ltl* LtlFalse::negClone() const
{ 
  return new LtlTrue(); 
}




/// ----------Methods for LtlTrue----------

const LtlGraph LtlTrue::expand1 (const LtlGraph::Node &node,
				 const LtlGraph &set) const
{ 
  LtlGraph::Node newnode(node);
  if (newnode.m_atomic.isEmpty())
    newnode.m_atomic = newnode.m_atomic + LtlFormula(*this);
  newnode.m_old    = newnode.m_old + LtlFormula(*this);
  return expand (newnode, set);
}




/// ----------Methods for LtlAtom----------

const LtlGraph LtlAtom::expand1 (const LtlGraph::Node &node,
				 const LtlGraph &set) const
{   
  LtlGraph::Node newnode;

  if (node.m_old >= LtlFormula(LtlNegAtom(m_value)))
    return set;
  else {
    newnode=node;
    if (newnode.m_atomic==Set<LtlFormula>(LtlFormula(LtlTrue())))
      newnode.m_atomic = Set<LtlFormula>(LtlFormula(*this));
    else
      newnode.m_atomic = newnode.m_atomic + LtlFormula(*this);
    newnode.m_old    = newnode.m_old + LtlFormula(*this);
    return expand (newnode, set);
  }
}


Ltl* LtlAtom::negClone() const
{ 
  return new LtlNegAtom(m_value); 
}




/// ----------Methods for LtlNegAtom----------

const LtlGraph LtlNegAtom::expand1 (const LtlGraph::Node &node,
				    const LtlGraph &set) const
{   
  LtlGraph::Node newnode;

  if (node.m_old >= LtlFormula(LtlAtom(m_value)))
    return set;
  else {
    newnode=node;
    if (newnode.m_atomic==Set<LtlFormula>(LtlFormula(LtlTrue())))
      newnode.m_atomic = Set<LtlFormula>(LtlFormula(*this));
    else
      newnode.m_atomic = newnode.m_atomic + LtlFormula(*this);
    newnode.m_old    = newnode.m_old + LtlFormula(*this);
    return expand (newnode, set);
  }
}





/// ----------Methods for LtlAnd----------

const LtlGraph LtlAnd::expand1 (const LtlGraph::Node &node,
				const LtlGraph &set) const
{
  LtlFormula mu(*m_left), psi(*m_right), eta(*this);

  return expand(LtlGraph::Node(node.m_name,
			       node.m_father,
			       node.m_incoming,
			       node.m_new + ((Set<LtlFormula>(mu) + psi) -
					     node.m_old),
			       node.m_old + eta,
			       node.m_atomic,
			       node.m_next),
		set);
}


LtlAnd::LtlAnd(Ltl *lf, Ltl *rf)
{
  if (lf->expr()<rf->expr()) {
    m_left=lf;
    m_right=rf;
  } else {
    m_left=rf;
    m_right=lf;
  }
}


Ltl* LtlAnd::negClone() const
{ 
  // Note that the twin constructor is used instead of the public
  // constructor. This binds the created negClones directly to the
  // LtlOr formula. No copies are left hanging.
  return new LtlOr(m_left->negClone(), m_right->negClone()); 
}


LtlAnd::LtlAnd(const Ltl &lf, const Ltl &rf)
{
  if (lf.expr()<rf.expr()) {
    m_left=lf.clone();
    m_right=rf.clone();
  } else {
    m_left=rf.clone();
    m_right=lf.clone();
  }
}


LtlAnd::~LtlAnd()
{
  delete m_left;
  delete m_right;
}




/// ----------Methods for LtlOr----------

const LtlGraph LtlOr::expand1 (const LtlGraph::Node &node,
			       const LtlGraph &set) const
{
  LtlFormula     mu(*m_left), psi(*m_right), eta(*this);
  LtlGraph::Node node1, node2;

  node1=LtlGraph::Node(newName(),
		       node.m_name,
		       node.m_incoming,
		       node.m_new + (Set<LtlFormula>(mu) - node.m_old),
		       node.m_old + eta,
		       node.m_atomic,
		       node.m_next);
  node2=LtlGraph::Node(newName(),
		       node.m_name,
		       node.m_incoming,
		       node.m_new + (Set<LtlFormula>(psi) - node.m_old),
		       node.m_old + eta,
		       node.m_atomic,
		       node.m_next);
  return expand(node2, expand(node1, set));
}


LtlOr::LtlOr(Ltl *lf, Ltl *rf)
{
  if (lf->expr()<rf->expr()) {
    m_left=lf;
    m_right=rf;
  } else {
    m_left=rf;
    m_right=lf;
  }
}


Ltl* LtlOr::negClone() const
{ 
  // Note that the twin constructor is used instead of the public
  // constructor. This binds the created negClones directly to the
  // LtlAnd formula. No copies are left hanging.
  return new LtlAnd(m_left->negClone(), m_right->negClone()); 
}


LtlOr::LtlOr(const Ltl &lf, const Ltl &rf)
{
  if (lf.expr()<rf.expr()) {
    m_left=lf.clone();
    m_right=rf.clone();
  } else {
    m_left=rf.clone();
    m_right=lf.clone();
  }
}


LtlOr::~LtlOr()
{
  delete m_left;
  delete m_right;
}





/// ----------Methods for LtlUntil----------

const LtlGraph LtlUntil::expand1 (const LtlGraph::Node &node,
				  const LtlGraph &set) const
{
  LtlFormula     mu(*m_left), psi(*m_right), eta(*this);
  LtlGraph::Node node1, node2;

  node1=LtlGraph::Node(newName(),
		       node.m_name,
		       node.m_incoming,
		       node.m_new + (Set<LtlFormula>(mu) - node.m_old),
		       node.m_old + eta,
		       node.m_atomic,
		       node.m_next + eta);
  node2=LtlGraph::Node(newName(),
		       node.m_name,
		       node.m_incoming,
		       node.m_new + (Set<LtlFormula>(psi) - node.m_old),
		       node.m_old + eta,
		       node.m_atomic,
		       node.m_next);
  return expand(node2, expand(node1, set));
}


LtlUntil::LtlUntil(Ltl *lf, Ltl *rf)
{
  m_left=lf;
  m_right=rf;
}


Ltl* LtlUntil::negClone() const
{ 
  // Note that the twin constructor is used instead of the public
  // constructor. This binds the created negClones directly to the
  // LtlResist formula. No copies are left hanging.
  return new LtlResist(m_left->negClone(), m_right->negClone()); 
}


LtlUntil::LtlUntil(const Ltl &lf, const Ltl &rf)
{
  m_left=lf.clone();
  m_right=rf.clone();
}


LtlUntil::~LtlUntil()
{
  delete m_left;
  delete m_right;
}




/// ----------Methods for LtlResist----------

const LtlGraph LtlResist::expand1 (const LtlGraph::Node &node,
			       const LtlGraph &set) const
{
  LtlFormula     mu(*m_left), psi(*m_right), eta(*this);
  LtlGraph::Node node1, node2;

  node1=LtlGraph::Node(newName(),
		       node.m_name,
		       node.m_incoming,
		       node.m_new + (Set<LtlFormula>(psi) - node.m_old),
		       node.m_old + eta,
		       node.m_atomic,
		       node.m_next + eta);
  node2=LtlGraph::Node(newName(),
		       node.m_name,
		       node.m_incoming,
		       node.m_new + ((Set<LtlFormula>(mu) + psi)- node.m_old),
		       node.m_old + eta,
		       node.m_atomic,
		       node.m_next);
  return expand(node2, expand(node1, set));
}


LtlResist::LtlResist(Ltl *lf, Ltl *rf)
{
  m_left=lf;
  m_right=rf;
}


Ltl* LtlResist::negClone() const
{ 
  // Note that the twin constructor is used instead of the public
  // constructor. This binds the created negClones directly to the
  // LtlResist formula. No copies are left hanging.
  return new LtlUntil(m_left->negClone(), m_right->negClone()); 
}


LtlResist::LtlResist(const Ltl &lf, const Ltl &rf)
{
  m_left=lf.clone();
  m_right=rf.clone();
}


LtlResist::~LtlResist()
{
  delete m_left;
  delete m_right;
}





/// ----------Methods for LtlNext----------

const LtlGraph LtlNext::expand1 (const LtlGraph::Node &node,
				 const LtlGraph &set) const
{
  LtlFormula mu(*m_formula), eta(*this);

  return expand(LtlGraph::Node(node.m_name,
			       node.m_father,
			       node.m_incoming,
			       node.m_new,
			       node.m_old + eta,
			       node.m_atomic,
			       node.m_next + mu),
		set);
}








